import logging

import click
from six.moves import range

from . import __version__, command


def _parse_host_port(addr):
    host, port = addr.split(':')
    return host, int(port)


@click.group(help='Note: each `--xxxx-addr` argument in the following commands'
             ' is in the form of HOST:PORT')
def cli():
    pass


@cli.command(help='Create a cluster with several Redis nodes')
@click.option(
    '--max-slots',
    type=int,
    default=1024,
    help='maxium number of slots in a single CLUSTER ADDSLOTS command')
@click.argument('addrs', nargs=-1, required=True)
def create(addrs, max_slots=1024):
    command.create([_parse_host_port(hp) for hp in addrs], max_slots)


@cli.command(help='Add a new Redis node to a cluster')
@click.option(
    '--existing-addr',
    required=True,
    help='Address of any node in the cluster')
@click.option('--new-addr', required=True, help='Address of the new node')
def add_node(existing_addr, new_addr):
    cluster_host, cluster_port = _parse_host_port(existing_addr)
    newin_host, newin_port = _parse_host_port(new_addr)
    command.add_node(cluster_host, cluster_port, newin_host, newin_port)


@cli.command(help='Add a slave node to a master')
@click.option('--master-addr', required=True, help='Address of the master')
@click.option('--slave-addr', required=True, help='Address of the slave')
def replicate(master_addr, slave_addr):
    master_host, master_port = _parse_host_port(master_addr)
    slave_host, slave_port = _parse_host_port(slave_addr)
    command.replicate(master_host, master_port, slave_host, slave_port)


@cli.command(help='Remove a Redis node from a cluster')
@click.option('--addr', required=True, help='Address of the node')
def del_node(addr):
    host, port = _parse_host_port(addr)
    command.del_node(host, port)


@cli.command(help='Shutdown a cluster. The cluster should have no more than'
             ' one Redis node and there should be no key in that Redis')
@click.option('--addr', required=True, help='Address of the node')
@click.option(
    '--ignore-failed',
    is_flag=True,
    default=False,
    help='Shutdown even if there are other nodes known as "failed"')
def shutdown(addr, ignore_failed):
    host, port = _parse_host_port(addr)
    command.shutdown_cluster(host, port, ignore_failed)


@cli.command(help='Fix migrating status')
@click.option('--addr', required=True, help='Address of the node')
def fix(addr):
    host, port = _parse_host_port(addr)
    command.fix_migrating(host, port)


@cli.command(help='Add a Redis node to a broken cluster to undertake missing'
             ' slots and recover the cluster status')
@click.option(
    '--existing-addr',
    required=True,
    help='Address of any node in the cluster')
@click.option('--new-addr', required=True, help='Address of the new node')
@click.option(
    '--max-slots',
    type=int,
    default=1024,
    help='maxium number of slots in a single CLUSTER ADDSLOTS command')
def rescue(existing_addr, new_addr, max_slots=1024):
    host, port = _parse_host_port(existing_addr)
    subst_host, subst_port = _parse_host_port(new_addr)
    command.rescue_cluster(host, port, subst_host, subst_port, max_slots)


@cli.command(help='Migrate slots from one Redis node to another')
@click.option(
    '--src-addr', required=True, help='Address of the migrating source')
@click.option(
    '--dst-addr', required=True, help='Address of the migrating destination')
@click.argument('slots_ranges', nargs=-1, required=True)
def migrate(src_addr, dst_addr, slots_ranges):
    src_host, src_port = _parse_host_port(src_addr)
    dst_host, dst_port = _parse_host_port(dst_addr)

    slots = []
    for rg in slots_ranges:
        if '-' in rg:
            begin, end = rg.split('-')
            slots.extend(range(int(begin), int(end) + 1))
        else:
            slots.append(int(rg))

    command.migrate_slots(src_host, src_port, dst_host, dst_port, slots)


def _format_master(node):
    s = 'M  %s %s %d' % (node.addr(), ','.join(node.flags),
                         len(node.assigned_slots))
    if node.slots_migrating:
        s += ' [migrating]'
    return s


def _format_slave(node, master):
    return ' S %s %s %s' % (node.addr(), ','.join(node.flags), master.addr())


@cli.command(help='List Redis nodes in a cluster')
@click.option(
    '--addr', required=True, help='Address of any node in the cluster')
def list(addr):
    host, port = _parse_host_port(addr)
    id_map = {}
    nodes = sorted(command.list_nodes(host, port)[0], key=lambda n: n.addr())
    master_count = 0
    fail_count = 0
    for node in nodes:
        node.slaves = []
        id_map[node.node_id] = node
        if node.fail:
            fail_count += 1
    for node in nodes:
        if node.slave:
            if node.master_id:
                id_map[node.master_id].slaves.append(node)
        else:
            master_count += 1
    print('Total %d nodes, %d masters, %d fail' % (len(nodes), master_count,
                                                   fail_count))
    for node in nodes:
        if node.master:
            print(_format_master(node))
            for slave in node.slaves:
                print(_format_slave(slave, node))


@cli.command(help='Send a command to all nodes in the cluster')
@click.option(
    '--master-only',
    is_flag=True,
    help='Only send to masters, and ignore --slave flag')
@click.option('--slave-only', is_flag=True, help='Only send to slaves')
@click.option(
    '--addr', required=True, help='Address of any node in the cluster')
@click.argument('commands', nargs=-1, required=True)
def execute(master_only, slave_only, addr, commands):
    host, port = _parse_host_port(addr)
    for r in command.execute(host, port, master_only, slave_only, commands):
        if r['result'] is None:
            print('%s -%s' % (r['node'].addr(), r['exception']))
        else:
            print('%s +%s' % (r['node'].addr(), r['result']))


def main():
    logging.basicConfig(level=logging.INFO)
    click.echo('Redis-trib %s Copyright (c) HunanTV Platform developers' %
               __version__)
    cli()



from os import environ

from ltcli import config


def get_cli_prompt():
    """Return cli prompt

    :param user: user name
    :return: prompt string
    """
    cluster_id = config.get_cur_cluster_id(allow_empty_id=True)
    if cluster_id < 0:
        cluster_id = '-'
    user = environ['USER']
    prompt_classname = u'class:ansiwhite'
    prompt_message = u'{}@lightningdb:{}> '.format(user, cluster_id)
    return [(prompt_classname, prompt_message)]


def get_sql_prompt():
    """Return sql prompt

    :param user: user name
    :return: prompt string
    """
    cluster_id = config.get_cur_cluster_id(allow_empty_id=True)
    if cluster_id < 0:
        cluster_id = '-'
    user = environ['USER']
    prompt_classname = u'class:ansigreen'
    prompt_message = u'({}){}@lightningdb:sql> '.format(cluster_id, user)
    return [(prompt_classname, prompt_message)]



class RedisErrorBase(Exception):
    def __init__(self, message, host, port):
        Exception.__init__(self, '%s:%d - %s' % (host, port, message))
        self.host = host
        self.port = port


class RedisStatusError(RedisErrorBase):
    pass


class RedisIOError(IOError, RedisErrorBase):
    def __init__(self, error, host, port):
        IOError.__init__(self, error)
        RedisErrorBase.__init__(self, error, host, port)



from __future__ import print_function
from __future__ import absolute_import

import os
import time
import shutil
import socket

import click
import fire
from fire.core import FireExit
from prompt_toolkit import PromptSession
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.history import FileHistory
from prompt_toolkit.lexers import PygmentsLexer
from pygments.lexers.sql import SqlLexer
import paramiko
import yaml

from ltcli import log
from ltcli import net
from ltcli import config
from ltcli import utils
from ltcli import prompt
from ltcli import color
from ltcli import ask_util
from ltcli import cluster_util
from ltcli import editor
from ltcli.log import logger
from ltcli.cli import Cli
from ltcli.cluster import Cluster
from ltcli.center import Center
from ltcli.conf import Conf
from ltcli.thriftserver import ThriftServer
from ltcli.deploy_util import DeployUtil, DEPLOYED, PENDING
from ltcli.rediscli import RedisCliConfig
from ltcli.exceptions import (
    SSHConnectionError,
    HostConnectionError,
    HostNameError,
    FileNotExistError,
    YamlSyntaxError,
    PropsSyntaxError,
    PropsKeyError,
    PropsError,
    ClusterIdError,
    ClusterNotExistError,
    ClusterRedisError,
    LightningDBError,
    SSHCommandError,
    EnvError,
)


user_info = {
    'user': None,
    'print_mode': 'screen'
}


def run_monitor(n=10):
    """Run monitor command

    Monitor remote logs
    """
    host_list = config.get_master_host_list()
    cluster_id = config.get_cur_cluster_id()
    target_host = ask_util.host_for_monitor(host_list)
    logger.info('Press Ctrl-C for exit.')
    client = net.get_ssh(target_host)
    path_of_fb = config.get_path_of_fb(cluster_id)
    sr2_redis_log = path_of_fb['sr2_redis_log']
    command = 'tail -F -n {} {}/servers*'.format(n, sr2_redis_log)
    net.ssh_execute_async(client, command)


# def run_deploy_v3(cluster_id=None, history_save=True, force=False):
def run_deploy(
        cluster_id=None,
        history_save=True,
        clean=False,
        strategy="none"
):
    # validate cluster id
    if cluster_id is None:
        cluster_id = config.get_cur_cluster_id(allow_empty_id=True)
        if cluster_id < 0:
            msg = 'Select cluster first or type cluster id with argument'
            logger.error(msg)
            return
    if not cluster_util.validate_id(cluster_id):
        raise ClusterIdError(cluster_id)

    # validate option
    if not isinstance(history_save, bool):
        logger.error("option '--history-save' can use only 'True' or 'False'")
        return
    logger.debug("option '--history-save': {}".format(history_save))
    if not isinstance(clean, bool):
        logger.error("option '--clean' can use only 'True' or 'False'")
        return
    logger.debug("option '--clean': {}".format(clean))
    # if not isinstance(force, bool):
    #     logger.error("option '--force' can use only 'True' or 'False'")
    #     return
    # logger.debug("option '--force': {}".format(force))
    strategy_list = ["none", "zero-downtime"]
    if strategy not in strategy_list:
        logger.error("DeployStrategyError: '{}'. Select in {}".format(
            strategy,
            strategy_list
        ))
        return
    if strategy == "zero-downtime":
        _deploy_zero_downtime(cluster_id)
        return
    _deploy(cluster_id, history_save, clean)


def _deploy_zero_downtime(cluster_id):
    logger.debug("zero downtime update cluster {}".format(cluster_id))
    center = Center()
    center.update_ip_port()
    m_hosts = center.master_host_list
    m_ports = center.master_port_list
    s_hosts = center.slave_host_list
    s_ports = center.slave_port_list
    path_of_fb = config.get_path_of_fb(cluster_id)

    # check master alive
    m_count = len(m_hosts) * len(m_ports)
    alive_m_count = center.get_alive_master_redis_count()
    if alive_m_count < m_count:
        logger.error("There is disconnected master")
        return

    if not config.is_slave_enabled:
        logger.error("Need to slave")
        return

    # select installer
    installer_path = ask_util.installer()
    installer_name = os.path.basename(installer_path)

    # backup info
    current_time = time.strftime("%Y%m%d%H%M%S", time.gmtime())
    conf_backup_dir = 'cluster_{}_conf_bak_{}'.format(cluster_id, current_time)
    cluster_backup_dir = 'cluster_{}_bak_{}'.format(cluster_id, current_time)
    local_ip = config.get_local_ip()

    # backup conf
    center.conf_backup(local_ip, cluster_id, conf_backup_dir)

    # backup cluster
    for host in s_hosts:
        cluster_path = path_of_fb['cluster_path']
        client = net.get_ssh(host)
        center.cluster_backup(host, cluster_id, cluster_backup_dir)
        client.close()

    # transfer & install
    logger.info('Transfer installer and execute...')
    for host in m_hosts:
        logger.info(' - {}'.format(host))
        client = net.get_ssh(host)
        cmd = 'mkdir -p {0} && touch {0}/.deploy.state'.format(cluster_path)
        net.ssh_execute(client=client, command=cmd)
        client.close()
        DeployUtil().transfer_installer(host, cluster_id, installer_path)
        try:
            DeployUtil().install(host, cluster_id, installer_name)
        except SSHCommandError as ex:
            msg = "Fail to execute installer '{}'".format(installer_path)
            logger.error(msg)
            logger.exception(ex)
            return

    # restore conf
    center.conf_restore(local_ip, cluster_id, conf_backup_dir)

    # set deploy state complete
    for node in m_hosts:
        path_of_fb = config.get_path_of_fb(cluster_id)
        cluster_path = path_of_fb['cluster_path']
        client = net.get_ssh(node)
        cmd = 'rm -rf {}'.format(os.path.join(cluster_path, '.deploy.state'))
        net.ssh_execute(client=client, command=cmd)
        client.close()

    # restart slave
    center.stop_redis(master=False)
    center.configure_redis(master=False)
    center.sync_conf()
    center.start_redis_process(master=False)
    center.wait_until_all_redis_process_up()

    # check slave is alive
    slaves_for_failover = center.check_all_master_have_alive_slave()

    key = 'cluster-node-timeout'
    origin_m_value = center.cli_config_get(key, m_hosts[0], m_ports[0])
    origin_s_value = center.cli_config_get(key, s_hosts[0], s_ports[0])
    logger.info('config set: cluster-node-timeout 2000')
    RedisCliConfig().set(key, '2000', all=True)

    # cluster failover (with no option)
    logger.info("Replace master to slave")
    logger.debug(slaves_for_failover)
    try_count = 0
    while try_count < 10:
        try_count += 1
        success = True
        for slave_addr in slaves_for_failover:
            host, port = slave_addr.split(':')
            # In some cases, the cluster failover is not complete
            # even if stdout is OK
            # If redis changed to master completely,
            # return 'ERR You should send CLUSTER FAILOVER to a slave'
            stdout = center.run_failover("{}:{}".format(host, port))
            logger.debug("failover {}:{} {}".format(host, port, stdout))
            if stdout != "ERR You should send CLUSTER FAILOVER to a slave":
                success = False
        if success:
            break
        logger.info("retry: {}".format(try_count))
        time.sleep(5)
    logger.info('restore config: cluster-node-timeout')
    center.cli_config_set_all(key, origin_m_value, m_hosts, m_ports)
    center.cli_config_set_all(key, origin_s_value, s_hosts, s_ports)
    if not success:
        logger.error("Fail to cluster failover")
        return

    # restart master (current slave)
    center.stop_redis(slave=False)
    center.configure_redis(slave=False)
    center.sync_conf()
    center.start_redis_process(slave=False)
    center.wait_until_all_redis_process_up()

    # change host info of redis.properties
    props_path = path_of_fb['redis_properties']
    after_m_ports = list(set(map(
        lambda x: int(x.split(':')[1]),
        slaves_for_failover
    )))
    after_s_ports = list(set(s_ports + m_ports) - set(after_m_ports))
    logger.debug("master port {}".format(m_ports))
    logger.debug("slave port {}".format(s_ports))
    key = 'sr2_redis_master_ports'
    logger.debug("after master port {}".format(after_m_ports))
    value = cluster_util.convert_list_2_seq(after_m_ports)
    logger.debug("converted {}".format(value))
    config.set_props(props_path, key, value)
    key = 'sr2_redis_slave_ports'
    logger.debug("after slave port {}".format(after_s_ports))
    value = cluster_util.convert_list_2_seq(after_s_ports)
    logger.debug("converted {}".format(value))
    config.set_props(props_path, key, value)


def _deploy(cluster_id, history_save, clean):
    deploy_state = DeployUtil().get_state(cluster_id)
    if deploy_state == DEPLOYED:
        q = [
            color.YELLOW,
            '(Watch out) ',
            'Cluster {} is already deployed. '.format(cluster_id),
            'Do you want to deploy again?',
            color.ENDC,
        ]
        yes = ask_util.askBool(''.join(q), default='n')
        if not yes:
            logger.info('Cancel deploy.')
            return

    restore_yes = None
    current_time = time.strftime("%Y%m%d%H%M%S", time.gmtime())
    cluster_backup_dir = 'cluster_{}_bak_{}'.format(cluster_id, current_time)
    conf_backup_dir = 'cluster_{}_conf_bak_{}'.format(cluster_id, current_time)
    tmp_backup_dir = 'cluster_{}_conf_bak_{}'.format(cluster_id, 'tmp')
    meta = [['NAME', 'VALUE']]
    path_of_fb = config.get_path_of_fb(cluster_id)
    conf_path = path_of_fb['conf_path']
    props_path = path_of_fb['redis_properties']
    cluster_path = path_of_fb['cluster_path']
    path_of_cli = config.get_path_of_cli(cluster_id)
    conf_backup_path = path_of_cli['conf_backup_path']
    tmp_backup_path = os.path.join(conf_backup_path, tmp_backup_dir)
    local_ip = config.get_local_ip()

    # ask installer
    installer_path = ask_util.installer()
    installer_name = os.path.basename(installer_path)
    meta.append(['installer', installer_name])

    # ask restore conf
    if deploy_state == DEPLOYED:
        restore_yes = ask_util.askBool('Do you want to restore conf?')
        meta.append(['restore', restore_yes])

    # input props
    hosts = []
    if deploy_state == DEPLOYED:
        if restore_yes:
            meta += DeployUtil().get_meta_from_props(props_path)
            hosts = config.get_props(props_path, 'sr2_redis_master_hosts')
        else:
            if os.path.exists(tmp_backup_path):
                q = 'There is a history of modification. Do you want to load?'
                yes = ask_util.askBool(q)
                if not yes:
                    shutil.rmtree(tmp_backup_path)
            if not os.path.exists(tmp_backup_path):
                shutil.copytree(conf_path, tmp_backup_path)
            tmp_props_path = os.path.join(tmp_backup_path, 'redis.properties')
            editor.edit(tmp_props_path, syntax='sh')
            meta += DeployUtil().get_meta_from_props(tmp_props_path)
            hosts = config.get_props(tmp_props_path, 'sr2_redis_master_hosts')
    else:
        # new deploy
        props_dict = ask_util.props(cluster_id, save=history_save)
        hosts = props_dict['hosts']
        meta += DeployUtil().get_meta_from_dict(props_dict)
    utils.print_table(meta)

    msg = [
        'Do you want to proceed with the deploy ',
        'accroding to the above information?',
    ]
    yes = ask_util.askBool(''.join(msg))
    if not yes:
        logger.info("Cancel deploy.")
        return

    # check node status
    success = Center().check_hosts_connection(hosts, True)
    if not success:
        logger.error('There are unavailable host.')
        return
    logger.debug('Connection of all hosts ok.')
    success = Center().check_include_localhost(hosts)
    if not success:
        logger.error('Must include localhost.')
        return

    # if pending, delete legacy on each hosts
    for host in hosts:
        if DeployUtil().get_state(cluster_id, host) == PENDING:
            client = net.get_ssh(host)
            command = 'rm -rf {}'.format(cluster_path)
            net.ssh_execute(client=client, command=command)
            client.close()

    # added_hosts = post_hosts - pre_hosts
    logger.info('Checking for cluster exist...')
    added_hosts = set(hosts)
    meta = []
    if deploy_state == DEPLOYED:
        pre_hosts = config.get_props(props_path, 'sr2_redis_master_hosts')
        added_hosts -= set(pre_hosts)
    can_deploy = True
    for host in added_hosts:
        client = net.get_ssh(host)
        if net.is_exist(client, cluster_path):
            meta.append([host, color.red('CLUSTER EXIST')])
            can_deploy = False
            continue
        meta.append([host, color.green('CLEAN')])
    if meta:
        utils.print_table([['HOST', 'STATUS']] + meta)
    if not can_deploy:
        logger.error('Cluster information exist on some hosts.')
        return
        # if not force:
        #     logger.error("If you trying to force, use option '--force'")
        #     return
    logger.info('OK')

    # cluster stop and clean
    if deploy_state == DEPLOYED:
        center = Center()
        cur_cluster_id = config.get_cur_cluster_id(allow_empty_id=True)
        run_cluster_use(cluster_id)
        center.update_ip_port()
        center.stop_redis()
        if clean:
            center.remove_all_of_redis_log_force()
            center.cluster_clean()
        run_cluster_use(cur_cluster_id)

    # backup conf
    if deploy_state == DEPLOYED:
        Center().conf_backup(local_ip, cluster_id, conf_backup_dir)

    # backup cluster
    backup_hosts = []
    if deploy_state == DEPLOYED:
        backup_hosts += set(pre_hosts)
    # if force:
    #     backup_hosts += added_hosts
    for host in backup_hosts:
        cluster_path = path_of_fb['cluster_path']
        client = net.get_ssh(host)
        Center().cluster_backup(host, cluster_id, cluster_backup_dir)
        client.close()

    # transfer & install
    logger.info('Transfer installer and execute...')
    for host in hosts:
        logger.info(' - {}'.format(host))
        client = net.get_ssh(host)
        cmd = 'mkdir -p {0} && touch {0}/.deploy.state'.format(cluster_path)
        net.ssh_execute(client=client, command=cmd)
        client.close()
        DeployUtil().transfer_installer(host, cluster_id, installer_path)
        try:
            DeployUtil().install(host, cluster_id, installer_name)
        except SSHCommandError as ex:
            msg = "Fail to execute installer '{}'".format(installer_path)
            logger.error(msg)
            logger.exception(ex)
            return

    # setup props
    if deploy_state == DEPLOYED:
        if restore_yes:
            tag = conf_backup_dir
        else:
            tag = tmp_backup_dir
        Center().conf_restore(local_ip, cluster_id, tag)
    else:
        key = 'sr2_redis_master_hosts'
        config.make_key_enable(props_path, key)
        config.set_props(props_path, key, props_dict['hosts'])

        key = 'sr2_redis_master_ports'
        config.make_key_enable(props_path, key)
        value = cluster_util.convert_list_2_seq(props_dict['master_ports'])
        config.set_props(props_path, key, value)

        key = 'sr2_redis_slave_hosts'
        config.make_key_enable(props_path, key)
        config.set_props(props_path, key, props_dict['hosts'])
        config.make_key_disable(props_path, key)

        if props_dict['replicas'] > 0:
            key = 'sr2_redis_slave_hosts'
            config.make_key_enable(props_path, key)

            key = 'sr2_redis_slave_ports'
            config.make_key_enable(props_path, key)
            value = cluster_util.convert_list_2_seq(props_dict['slave_ports'])
            config.set_props(props_path, key, value)

        key = 'ssd_count'
        config.make_key_enable(props_path, key)
        config.set_props(props_path, key, props_dict['ssd_count'])

        key = 'sr2_redis_data'
        config.make_key_enable(props_path, key, v1_flg=True)
        config.make_key_enable(props_path, key, v1_flg=True)
        config.make_key_disable(props_path, key)
        config.set_props(props_path, key, props_dict['prefix_of_db_path'])

        key = 'sr2_redis_db_path'
        config.make_key_enable(props_path, key, v1_flg=True)
        config.make_key_enable(props_path, key, v1_flg=True)
        config.make_key_disable(props_path, key)
        config.set_props(props_path, key, props_dict['prefix_of_db_path'])

        key = 'sr2_flash_db_path'
        config.make_key_enable(props_path, key, v1_flg=True)
        config.make_key_enable(props_path, key, v1_flg=True)
        config.make_key_disable(props_path, key)
        config.set_props(props_path, key, props_dict['prefix_of_db_path'])

    # synk props
    logger.info('Sync conf...')
    for node in hosts:
        if socket.gethostbyname(node) in config.get_local_ip_list():
            continue
        client = net.get_ssh(node)
        if not client:
            logger.error("ssh connection fail: '{}'".format(node))
            return
        net.copy_dir_to_remote(client, conf_path, conf_path)
        client.close()

    # set deploy state complete
    if os.path.exists(tmp_backup_path):
        shutil.rmtree(tmp_backup_path)
    for node in hosts:
        path_of_fb = config.get_path_of_fb(cluster_id)
        cluster_path = path_of_fb['cluster_path']
        client = net.get_ssh(node)
        cmd = 'rm -rf {}'.format(os.path.join(cluster_path, '.deploy.state'))
        net.ssh_execute(client=client, command=cmd)
        client.close()

    logger.info('Complete to deploy cluster {}.'.format(cluster_id))
    Cluster().use(cluster_id)


def run_cluster_use(cluster_id):
    print_mode = user_info['print_mode']
    c = Cluster(print_mode)
    c.use(cluster_id)


def run_import_conf():
    def _to_config_yaml(
          cluster_id, release, nodes, master_start_port, master_end_port,
          master_enabled, slave_start_port, slave_end_port, slave_enabled,
          ssd_count):
        conf = {}
        conf['release'] = release
        conf['nodes'] = nodes
        conf['ssd'] = {}
        conf['master_ports'] = {}
        conf['slave_ports'] = {}
        conf['master_ports']['from'] = int(master_start_port)
        conf['master_ports']['to'] = int(master_end_port)
        conf['master_ports']['enabled'] = bool(master_enabled)
        conf['slave_ports']['from'] = int(slave_start_port)
        conf['slave_ports']['to'] = int(slave_end_port)
        conf['slave_ports']['enabled'] = bool(slave_enabled)
        conf['ssd']['count'] = int(ssd_count)

        root_of_cli_config = config.get_root_of_cli_config()
        cluster_base_path = os.path.join(root_of_cli_config, 'clusters')
        if not os.path.isdir(cluster_base_path):
            os.mkdir(cluster_base_path)
        cluster_path = os.path.join(root_of_cli_config, 'clusters', cluster_id)
        if not os.path.isdir(cluster_path):
            os.mkdir(cluster_path)
        yaml_path = os.path.join(cluster_path, 'config.yaml')
        with open(yaml_path, 'w') as fd:
            yaml.dump(conf, fd, default_flow_style=False)

    def _import_from_fb_to_cli_conf(rp_exists):
        for cluster_id in rp_exists:
            path_of_fb = config.get_path_of_fb(cluster_id)
            rp = path_of_fb['redis_properties']
            d = config.get_props_as_dict(rp)
            nodes = d['sr2_redis_master_hosts']
            master_start_port = 0
            master_end_port = 0
            slave_start_port = 0
            slave_end_port = 0
            master_enabled = 'sr2_redis_master_ports' in d
            slave_enabled = 'sr2_redis_slave_ports' in d
            if master_enabled:
                master_start_port = min(d['sr2_redis_master_ports'])
                master_end_port = max(d['sr2_redis_master_ports'])
            if slave_enabled:
                slave_start_port = min(d['sr2_redis_slave_ports'])
                slave_end_port = max(d['sr2_redis_slave_ports'])
            ssd_count = d['ssd_count']
            _to_config_yaml(
                cluster_id=cluster_id,
                release='',
                nodes=nodes,
                master_start_port=master_start_port,
                master_end_port=master_end_port,
                master_enabled=master_enabled,
                slave_start_port=slave_start_port,
                slave_end_port=slave_end_port,
                slave_enabled=slave_enabled,
                ssd_count=ssd_count)
            logger.info('Save config.yaml from redis.properties')

    def _get_cluster_ids_from_fb():
        cluster_id = config.get_cur_cluster_id(allow_empty_id=True)
        path_of_fb = config.get_path_of_fb(cluster_id)
        base_directory = path_of_fb['base_directory']
        dirs = [f for f in os.listdir(base_directory)
                if not os.path.isfile(os.path.join(base_directory, f))]
        cluster_ids = [d.split('_')[1] for d in dirs if 'cluster_' in d]
        return cluster_ids

    cluster_ids = _get_cluster_ids_from_fb()
    root_of_cli_config = config.get_root_of_cli_config()

    rp_exists = []
    rp_not_exists = []
    dest_folder_exists = []
    meta = [['cluster_id', 'state']]
    for cluster_id in cluster_ids:
        path_of_fb = config.get_path_of_fb(cluster_id)
        rp = path_of_fb['redis_properties']
        dest_path = os.path.join(root_of_cli_config, 'clusters', cluster_id)
        dest_path = os.path.join(dest_path, 'config.yaml')
        cluster_path = path_of_fb['cluster_path']
        deploy_state = os.path.join(cluster_path, '.deploy.state')
        if os.path.exists(dest_path):
            dest_folder_exists.append(cluster_id)
            meta.append([cluster_id, 'SKIP(dest_exist)'])
        elif os.path.isfile(rp) and not os.path.isfile(deploy_state):
            rp_exists.append(cluster_id)
            meta.append([cluster_id, 'IMPORT'])
        else:
            rp_not_exists.append(cluster_id)
            meta.append([cluster_id, 'SKIP(broken)'])

    logger.info('Diff fb and cli conf folders.')
    utils.print_table(meta)
    if rp_exists:
        return
    import_yes = ask_util.askBool('Do you want to import conf?', ['y', 'n'])
    if not import_yes:
        return

    _import_from_fb_to_cli_conf(rp_exists)


def run_exit():
    """EXit ltcli
    """
    # empty function for docs of fire
    pass


def run_clear():
    """Clear screen
    """
    # empty function for docs of fire
    pass


class Command(object):
    """This is LightningDB command line.
We use python-fire(https://github.com/google/python-fire)
for automatically generating CLIs

    - deploy: install LightningDB package to nodes
    - c: change cluster #, alias of cluster use
    - cluster: trib.rb cluster wrapper
    - cli: redis-cli command wrapper
    - conf: edit conf file
    - monitor: monitor redis logs
    - thriftserver: thriftserver command
    - ths: alias of thriftserver
    - ll: change log level to debug ltcli
    - exit: exit ltcli
    - clear: clear screen
 """

    def __init__(self):
        """Member variables will be cli
        """
        # pylint: disable=invalid-name
        # cli command naming is not have to follow snake_caes
        self.deploy = run_deploy
        self.c = run_cluster_use
        self.cluster = Cluster()
        self.cli = Cli()
        self.conf = Conf()
        self.monitor = run_monitor
        self.thriftserver = ThriftServer()
        self.ths = ThriftServer()
        self.ll = log.set_level
        self.exit = run_exit
        self.clear = run_clear


def _handle(text):
    if text == '':
        return
    if text == 'clear':
        utils.clear_screen()
        return
    text = text.replace('-- --help', '?')
    text = text.replace('--help', '?')
    text = text.replace('?', '-- --help')
    try:
        fire.Fire(
            component=Command,
            command=text)
    except KeyboardInterrupt:
        logger.warning('\b\bCanceled')
    except KeyError as ex:
        logger.warn('[%s] command fail' % text)
        logger.exception(ex)
    except TypeError as ex:
        logger.exception(ex)
    except IOError as ex:
        if ex.errno == 2:
            logger.error("{}: '{}'".format('FileNotExistError', ex.filename))
        else:
            logger.exception(ex)
    except EOFError:
        logger.warning('\b\bCanceled')
    except utils.CommandError as ex:
        logger.exception(ex)
    except FireExit as ex:
        pass
    except (
            HostNameError,
            HostConnectionError,
            SSHConnectionError,
            FileNotExistError,
            YamlSyntaxError,
            PropsSyntaxError,
            PropsKeyError,
            PropsError,
            SSHCommandError,
            ClusterRedisError,
            ClusterNotExistError,
            ClusterIdError,
            EnvError,
    ) as ex:
        logger.error('{}: {}'.format(ex.class_name(), str(ex)))
    except LightningDBError as ex:
        logger.error('[ErrorCode {}] {}'.format(ex.error_code, str(ex)))
    except BaseException as ex:
        logger.exception(ex)


def _initial_check():
    try:
        # Simple check to see if ssh access to localhost is possible
        net.get_ssh('localhost')
    except paramiko.ssh_exception.SSHException:
        logger.error('Need to ssh-keygen for localhost')
        exit(1)
    cli_config = config.get_cli_config()
    try:
        base_directory = cli_config['base_directory']
    except KeyError:
        pass
    except TypeError:
        root_of_cli_config = config.get_root_of_cli_config()
        conf_path = os.path.join(root_of_cli_config, 'config')
        os.system('rm {}'.format(conf_path))
        base_directory = None
    if not base_directory or not base_directory.startswith(('~', '/')):
        base_directory = ask_util.base_directory()
    base_directory = os.path.expanduser(base_directory)
    if not os.path.isdir(base_directory):
        os.system('mkdir -p {}'.format(base_directory))


def _validate_cluster_id(cluster_id):
    try:
        if cluster_id is None:
            cluster_id = config.get_cur_cluster_id(allow_empty_id=True)
        elif not utils.is_number(cluster_id):
            raise ClusterIdError(cluster_id)
        cluster_id = int(cluster_id)
        run_cluster_use(cluster_id)
        return cluster_id
    except (ClusterIdError, ClusterNotExistError) as ex:
        logger.warning(ex)
        cluster_id = -1
        run_cluster_use(cluster_id)
        return cluster_id


def print_version():
    here = os.path.abspath(os.path.dirname(__file__))
    about = {}
    with open(os.path.join(here, '__version__.py'), 'r') as f:
        exec(f.read(), about)
    version = about['__version__']
    print('ltcli version {}'.format(version))


@click.command()
@click.option('-c', '--cluster_id', default=None, help='ClusterId.')
@click.option('-d', '--debug', default=False, help='Debug.')
@click.option('-v', '--version', is_flag=True, help='Version.')
def main(cluster_id, debug, version):
    if version:
        print_version()
        return
    _initial_check()
    if debug:
        log.set_mode('debug')

    logger.debug('Start ltcli')

    cluster_id = _validate_cluster_id(cluster_id)

    history = os.path.join(config.get_root_of_cli_config(), 'cli_history')
    session = PromptSession(
        lexer=PygmentsLexer(SqlLexer),
        history=FileHistory(history),
        auto_suggest=AutoSuggestFromHistory(),
        style=utils.style)
    while True:
        try:
            p = prompt.get_cli_prompt()
            text = session.prompt(p, style=utils.style)
            if text == "exit":
                break
            if 'ltcli' in text:
                old = text
                text = text.replace('ltcli', '')
                logger.info('> You can use "%s" instead of "%s"' % (text, old))
            _handle(text)
        except KeyboardInterrupt:
            continue
        except EOFError:
            break


if __name__ == '__main__':
    # pylint: disable=no-value-for-parameter
    # Parameter used by Click
    main()



import os
import shutil


from ltcli import config
from ltcli.log import logger
from ltcli import editor
from ltcli.center import Center
from ltcli import ask_util


class Conf(object):
    """command for edit conf file.

    open conf file with editor.
    sync to all hosts after edit.
    """
    def __init__(self):
        pass

    def cluster(self):
        """edit 'redis.properties' of cluster
        """
        cluster_id = config.get_cur_cluster_id()
        path_of_fb = config.get_path_of_fb(cluster_id)
        target_path = path_of_fb['redis_properties']
        self._edit_conf(target_path, syntax='sh')
        center = Center()
        center.update_ip_port()
        success = center.check_hosts_connection()
        if not success:
            return
        success = center.sync_file(target_path)
        if success:
            logger.info('Complete edit')

    def master(self):
        """ edit 'redis-master.conf.template'
        """
        cluster_id = config.get_cur_cluster_id()
        path_of_fb = config.get_path_of_fb(cluster_id)
        target_path = path_of_fb['master_template']
        self._edit_conf(target_path, syntax='sh')
        center = Center()
        center.update_ip_port()
        success = center.check_hosts_connection()
        if not success:
            return
        success = center.sync_file(target_path)
        if success:
            logger.info('Complete edit')

    def slave(self):
        """edit 'redis-slave.conf.template'
        """
        cluster_id = config.get_cur_cluster_id()
        path_of_fb = config.get_path_of_fb(cluster_id)
        target_path = path_of_fb['slave_template']
        self._edit_conf(target_path, syntax='sh')
        center = Center()
        center.update_ip_port()
        success = center.check_hosts_connection()
        if not success:
            return
        success = center.sync_file(target_path)
        if success:
            logger.info('Complete edit')

    def thriftserver(self):
        """edit 'thriftserver.properties'
        """
        cluster_id = config.get_cur_cluster_id()
        path_of_fb = config.get_path_of_fb(cluster_id)
        target_path = path_of_fb['thrift_properties']
        self._edit_conf(target_path, syntax='sh')
        center = Center()
        center.update_ip_port()
        success = center.check_hosts_connection()
        if not success:
            return
        success = center.sync_file(target_path)
        if success:
            logger.info('Complete edit')

    def ths(self):
        """alias of thriftserver
        """
        self.thriftserver()

    def _edit_conf(self, target_path, syntax=None):
        tmp_target_path = target_path + '.tmp'
        if os.path.exists(tmp_target_path):
            q = 'There is a history of modification. Do you want to load?'
            yes = ask_util.askBool(q)
            if not yes:
                os.remove(tmp_target_path)
        if not os.path.exists(tmp_target_path):
            shutil.copy(target_path, tmp_target_path)
        editor.edit(tmp_target_path, syntax=syntax)
        shutil.copy(tmp_target_path, target_path)
        os.remove(tmp_target_path)



import logging
import random

from .command import custom_migrate_slots
from .custom_node import CustomClusterNode
from .custom_reshard import CustomReshard
from .custom_util import CustomStd

total_slot_count = 16384
migrate_default_timeout = 60000
rebalance_default_threshold = 2


class RedisTrib(object):
    def __init__(self, opt={}):
        self.opt = opt
        self.nodes = []
        self.fix = False
        self.errors = []
        self.timeout = migrate_default_timeout
        self.cluster_error = False
        self.master_nodes = []

    def add_node(self, node):
        self.nodes.append(node)

    def reset_node(self):
        self.nodes = []

    def load_cluster_info_from_node(self, node_addr):
        logging.debug('load_cluster_info_from_node')
        node = CustomClusterNode(node_addr)
        node.load_info()
        self.reset_node()
        self.add_node(node)
        for friend in node.friends:
            if 'noaddr' in friend['flags'] or \
                    'disconnected' in friend['flags'] or \
                    'fail' in friend['flags']:
                continue
            fnode = CustomClusterNode(friend['addr'])
            fnode.load_info()
            self.add_node(fnode)
        self.populate_nodes_replicas_info()
        for node in self.nodes:
            if 'master' in node.info['flags']:
                self.master_nodes.append(node)

    def populate_nodes_replicas_info(self):
        # TODO
        pass

    def rebalance_cluster_cmd(self):
        ip = self.opt['ip']
        port = self.opt['port']
        threshold = rebalance_default_threshold

        self.load_cluster_info_from_node('%s:%s' % (ip, port))
        self.check_cluster()
        master_nodes = self.master_nodes
        nodes_involved = len(master_nodes)
        total_weight = len(master_nodes)
        threshold_reached = False
        for node in master_nodes:
            i = node.info
            expected = int(float(total_slot_count) / total_weight)
            slot_count = len(i['slots'].keys())
            b = slot_count - expected
            i['balance'] = b
            over_threshold = False
            if threshold > 0:
                if slot_count > 0:
                    err_perc = abs(100 - (100.0 * expected / slot_count))
                    print('err_perc:', err_perc)
                    over_threshold = err_perc > threshold
                elif expected > 0:
                    over_threshold = True
            if over_threshold:
                threshold_reached = over_threshold
        if not threshold_reached:
            print('*** No rebalancing needed! '
                  'All nodes are within the %s%% threshold' % threshold)
            return

        total_balance = sum(node.info['balance'] for node in master_nodes)
        while total_balance > 0:
            for node in master_nodes:
                if node.info['balance'] < 0 and total_balance > 0:
                    node.info['balance'] -= 1
                    total_balance -= 1
        master_nodes.sort(key=lambda x: x.info['balance'])
        print('>>> Rebalancing across %s nodes. Total weight = %s' % (
            nodes_involved, total_weight))

        for node in master_nodes:
            b = node.info['balance']
            name = node.info['name']
            print('%s balance is %s' % (name, b))

        dst_idx = 0
        src_idx = len(master_nodes) - 1
        while dst_idx < src_idx:
            dst = master_nodes[dst_idx]
            src = master_nodes[src_idx]
            num_slots = min([abs(dst.info['balance']),
                                abs(src.info['balance']), ])
            if num_slots > 0:
                src_name = src.info['name']
                dst_name = dst.info['name']
                print('Moving %s slots from %s to %s' % (
                    num_slots, src_name, dst_name))
                self.move_slots(src, dst, num_slots)

            dst.info['balance'] += num_slots
            src.info['balance'] -= num_slots
            dst_up = dst.info['balance'] == 0
            src_down = src.info['balance'] == 0
            dst_idx += dst_up
            src_idx -= src_down

    def show_reshard_table(self, table):
        for row in table:
            print('    Moving slot %s from %s' % (
                row['slot'],
                row['source'].info['name']
            ))

    def move_slots(self, src, dst, num_slots, quiet=False):
        move_target_slots = self.compute_single_reshard_table(src, num_slots)
        with CustomStd(quiet=quiet):
            custom_migrate_slots(src, dst, move_target_slots)

    def compute_single_reshard_table(self, src, num_slots):
        moved = []
        slots = src.info['slots']
        t = slots.keys()
        t.sort()
        for slot_num in t[0:int(num_slots)]:
            moved.append(slot_num)
        return moved

    def name_to_node(self, name):
        for node in self.nodes:
            if node.info['name'] == name:
                return node
        assert False
        return None

    def get_node_by_name(self, name):
        return self.name_to_node(name)

    def compute_reshard_table(
        self,
        src_node_ids,
        move_slot_count):
        src_list = list(map(self.name_to_node, src_node_ids))
        moved = []
        src_list.sort(key=lambda x: len(x.info['slots']), reverse=True)
        total_slots = sum(len(node.info['slots']) for node in src_list)
        if total_slots < move_slot_count:
            assert False

        ratio = float(move_slot_count) / total_slots
        target_count_list = []
        for i, node in enumerate(src_list):
            slots = node.info['slots']
            cur_slot_count = len(slots)
            target_count = int(ratio * cur_slot_count)
            # add one more if possible
            target_count = min(target_count + 1, cur_slot_count)
            target_count_list.append(target_count)
        src_count = len(src_list)
        while True:
            cur_sum = sum(target_count_list)
            if cur_sum == move_slot_count:
                break
            elif cur_sum < move_slot_count:
                exit(1)
            else:
                idx = random.randrange(src_count)
                if target_count_list[idx] > 0:
                    target_count_list[idx] -= 1

        for i, target_count in enumerate(target_count_list):
            node = src_list[i]
            slots = node.info['slots']
            t = slots.keys()
            t.sort()
            slot_num_list = []
            for slot_num in t[0:int(target_count)]:
                slot_num_list.append(slot_num)
            moved.append({'node': node, 'slot_num_list': slot_num_list})
        return moved

    def reshard_cluster_cmd(self, slots, src_node_ids, dest_node_id, src_all):
        ip = self.opt['ip']
        port = self.opt['port']
        self.load_cluster_info_from_node('%s:%s' % (ip, port))
        self.check_cluster()
        trib = self
        config = {
            'total_slot_count': 16384,
        }
        CustomReshard(
            trib, slots, src_node_ids, dest_node_id, src_all, config).run()

    def show_nodes(self):
        for node in self.nodes:
            print(node.info_string())

    def check_config_consistency(self):
        if not self.is_config_consistent():
            print("[ERR] Nodes don't agree about configuration!")
            self.cluster_error = True
            return False
        else:
            print('[OK] All nodes agree about slots configuration.')
            return True

    def check_open_slots(self):
        print('>>> Check for open slots...')
        open_slots = []
        for node in self.nodes:
            m_nodes = node.info['migrating']
            i_nodes = node.info['importing']
            if len(m_nodes) > 0:
                self.cluster_error = True
                print('[WARNING] Node %s has lots in migrating state' %
                      node.to_string())
                open_slots.append(m_nodes.keys())
            elif len(i_nodes) > 0:
                print('[WARNING] Node %s has lots in importing state' %
                      node.to_string())
                open_slots.append(i_nodes.keys())
                # TODO: implement fix option

    def covered_slots(self):
        slots = {}
        for node in self.nodes:
            cur_slots = node.info['slots'].copy()
            slots.update(cur_slots)
        return slots.keys()

    def check_slots_coverage(self):
        print('>>> Check slots coverage...')
        if total_slot_count == len(self.covered_slots()):
            print('[OK] All %s slots covered' % total_slot_count)
        else:
            print('[ERR] Not all %s slots are covered by nodes.' %
                  total_slot_count)

    def is_config_consistent(self):
        prev = None
        for node in self.nodes:
            cur = node.get_config_signature()
            if prev is None:
                prev = cur
            else:
                if prev != cur:
                    return False
        return True

    def check_cluster(self, quiet=False):
        n = self.nodes[0]
        self.cluster_error = False
        if not quiet:
            self.show_nodes()
        print('>>> Performing Cluster Check (using node %s)' % n.to_string())
        self.check_config_consistency()
        self.check_open_slots()
        self.check_slots_coverage()
        if self.cluster_error:
            print('*** Please fix your cluster problems before execution')
            exit(1)

    def is_exist_master_node_id(self, node_id):
        for node in self.master_nodes:
            if node_id == node.info['name']:
                return True
        return False


def rebalance_cluster_cmd(ip, port):
    logging.debug('rebalance')
    rt = RedisTrib({'ip': ip, 'port': port})
    rt.rebalance_cluster_cmd()
    return True


def reshard_cluster_cmd(ip, port, slots=0, dest_node_id=None, src_all=False):
    logging.debug('reshard')
    rt = RedisTrib({'ip': ip, 'port': port})
    src_node_ids = []
    rt.reshard_cluster_cmd(slots, src_node_ids, dest_node_id, src_all)
    return True


def check_cluster_cmd(ip, port):
    logging.debug('check_cluster')
    rt = RedisTrib({'ip': ip, 'port': port})
    rt.load_cluster_info_from_node('%s:%s' % (ip, port))
    rt.check_cluster()
    return True



import os
import sys


class CustomStd(object):
    def __init__(self, quiet=True):
        self.devnull = open(os.devnull, 'w')
        if quiet:
            self._stdout = self.devnull or sys.stdout
            self._stderr = self.devnull or sys.stderr
        else:
            self._stdout = sys.stdout
            self._stderr = sys.stderr

    def __enter__(self):
        self.old_stdout, self.old_stderr = sys.stdout, sys.stderr
        self.old_stdout.flush()
        self.old_stderr.flush()
        sys.stdout, sys.stderr = self._stdout, self._stderr

    def __exit__(self, exc_type, exc_value, traceback):
        self._stdout.flush()
        self._stderr.flush()
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr
        self.devnull.close()


class PrettySlotGenerator(object):
    def __init__(self):
        self.pretty_list = []

    def generate(self, slots):
        key_list = slots.keys()
        key_list.sort()
        self.pretty_list = []
        buf = self._get_init_buf()
        for key in key_list:
            cur_state = buf['state']
            if cur_state == 'find_start':
                self._set_start_info(buf, key)
                continue
            elif cur_state == 'find_end':
                if self._is_continuous_number(buf['prev'], key):
                    buf['end'] = key
                    buf['prev'] = key
                    buf['count'] += 1
                else:
                    self._set_end_info(buf)
                    buf = self._get_init_buf()
                    self._set_start_info(buf, key)
            else:
                assert False
        return self.pretty_list

    def to_string(self, pretty_list=None):
        if pretty_list is None:
            pretty_list = self.pretty_list
        total = sum(int(v['count']) for v in pretty_list)
        msg = 'slots:'
        slot_str_list = []
        for item in pretty_list:
            s = item['start']
            e = item['end']
            slot_str_list.append('%d-%d' % (s, e))
        if len(slot_str_list) > 0:
            msg += ','.join(slot_str_list)
        msg += ' (%d slots)' % total
        return msg

    def _get_init_buf(self):
        buf = {
            'start': -1,
            'prev': -1,
            'end': -1,
            'count': 1,
            'state': 'find_start',
        }
        return buf

    def _set_start_info(self, buf, key):
        buf['state'] = 'find_end'
        buf['start'] = key
        buf['end'] = key
        buf['prev'] = key
        buf['count'] = 1
        self.pretty_list.append(buf)

    def _set_end_info(self, buf):
        buf['state'] = 'complete'
        buf['prev'] = -1

    def _is_continuous_number(self, prev, cur):
        diff = cur - prev
        return diff == 1



from werkzeug.utils import cached_property

from .connection import Connection


class ClusterNode(object):
    def __init__(self, node_id, latest_know_ip_address_and_port, flags,
                 master_id, last_ping_sent_time, last_pong_received_time,
                 node_index, link_status, *assigned_slots):
        self.node_id = node_id
        host, port = latest_know_ip_address_and_port.split('@')[0].split(':')
        self.host = host
        self.port = int(port)
        self.flags = flags.split(',')
        self.master_id = None if master_id == '-' else master_id
        self.assigned_slots = []
        self.slots_migrating = False
        for slots_range in assigned_slots:
            if '[' == slots_range[0] and ']' == slots_range[-1]:
                # exclude migrating slot
                self.slots_migrating = True
                continue
            if '-' in slots_range:
                begin, end = slots_range.split('-')
                self.assigned_slots.extend(range(int(begin), int(end) + 1))
            else:
                self.assigned_slots.append(int(slots_range))

        self._conn = None

    def addr(self):
        return '%s:%d' % (self.host, self.port)

    @cached_property
    def role_in_cluster(self):
        return 'master' if self.master else 'slave'

    @cached_property
    def myself(self):
        return 'myself' in self.flags

    @cached_property
    def master(self):
        return 'master' in self.flags

    @cached_property
    def slave(self):
        return not self.master

    @cached_property
    def fail(self):
        return 'fail' in self.flags or 'fail?' in self.flags

    def get_conn(self):
        if self._conn is None:
            self._conn = Connection(self.host, self.port)
        return self._conn

    def close(self):
        if self._conn is not None:
            self._conn.close()
            self._conn = None

    def talker(self):
        import warnings
        warnings.warn(
            'redistrib.clusternode.ClusterNode.talker is deprecated and will be removed at the next release.'
        )
        return self.get_conn()


class BaseBalancer(object):
    def weight(self, clusternode):
        return 1


def base_balance_plan(nodes, balancer=None):
    if balancer is None:
        balancer = BaseBalancer()
    nodes = [n for n in nodes if 'master' == n.role_in_cluster]
    origin_slots = [len(n.assigned_slots) for n in nodes]
    total_slots = sum(origin_slots)
    weights = [balancer.weight(n) for n in nodes]
    total_weight = sum(weights)

    result_slots = [total_slots * w // total_weight for w in weights]
    frag_slots = total_slots - sum(result_slots)

    migratings = [[n, r - o]
                  for n, r, o in zip(nodes, result_slots, origin_slots)]

    for m in migratings:
        if frag_slots > -m[1] > 0:
            frag_slots += m[1]
            m[1] = 0
        elif frag_slots <= -m[1]:
            m[1] += frag_slots
            break

    migrating = sorted(
        [m for m in migratings if m[1] != 0], key=lambda x: x[1])
    mig_out = 0
    mig_in = len(migrating) - 1

    plan = []
    while mig_out < mig_in:
        if migrating[mig_in][1] < -migrating[mig_out][1]:
            plan.append((migrating[mig_out][0], migrating[mig_in][0],
                         migrating[mig_in][1]))
            migrating[mig_out][1] += migrating[mig_in][1]
            mig_in -= 1
        elif migrating[mig_in][1] > -migrating[mig_out][1]:
            plan.append((migrating[mig_out][0], migrating[mig_in][0],
                         -migrating[mig_out][1]))
            migrating[mig_in][1] += migrating[mig_out][1]
            mig_out += 1
        else:
            plan.append((migrating[mig_out][0], migrating[mig_in][0],
                         migrating[mig_in][1]))
            mig_out += 1
            mig_in -= 1

    return plan



import os
import sys

from prompt_toolkit.styles import Style
from terminaltables import AsciiTable

from ltcli import config
from ltcli import editor
from ltcli.log import logger
from ltcli.exceptions import ConvertError


class TermColor:
    """This is for term text color
    """
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

    @staticmethod
    def green(msg):
        return TermColor.OKGREEN + msg + TermColor.ENDC

    @staticmethod
    def fail(msg):
        return TermColor.FAIL + msg + TermColor.ENDC

    @staticmethod
    def blue(msg):
        return TermColor.OKBLUE + msg + TermColor.ENDC


class RangeChecker(object):
    """This class for checking range of index

    Update min max value
    """
    def __init__(self):
        self.min_index = 99999999
        self.max_index = 0

    def check(self, cur_index):
        """Update min, max index value
        :param cur_index: cur index
        """
        if self.min_index > cur_index:
            self.min_index = cur_index
        if self.max_index < cur_index:
            self.max_index = cur_index


def print_table(meta):
    """Print data as table format
    """
    table = AsciiTable(meta)
    print(table.table)


def get_ip_port_tuple_list(ip_list, port_list):
    """Convert as ip | port tuple data(list)
    """
    targets = []
    for ip in ip_list:
        for port in port_list:
            targets.append((ip, port))
    return targets


def get_ip_port_dict_list(ip_list, port_list):
    """Convert as ip | port dict list
    """
    targets = []
    for ip in ip_list:
        for port in port_list:
            targets.append({'host': ip, 'port': port})
    return targets


class Error(Exception):
    """Base class for exceptions in this module."""
    pass


class DuplicatedError(Error):
    def __init__(self, target):
        self.target = target
        logger.error('DuplicatedError: %s' % target)


class CommandError(Error):
    def __init__(self, exit_status, command, hostname, port):
        self.exit_status = exit_status
        self.command = command
        self.hostname = hostname
        self.port = port
        msg = [
            'CommandError:',
            'exit_status={}'.format(exit_status),
            'command={}'.format(command),
            'host:port={}:{}'.format(hostname, port)
        ]
        logger.error('\n'.join(msg))


# table print
def tprint(dictionary, header=['key', 'value']):
    """Table print dictionary

    :param dictionary: dictionary
    :param header: header
    """
    items = dictionary.items()
    sorted_list = sorted(items, key=lambda x: x[1])
    tr = TableReport(header)
    tr.data = sorted_list
    tr.print_out()


def tprint_list(rows, header=['key', 'value1', 'value2']):
    """Table print list

    :param rows: list
    :param header: header
    """
    tr = TableReport(header)
    tr.data = rows
    tr.print_out()


class TableReport(object):
    """This is for logging important information and print out as table format
    """
    def __init__(self, column):
        self.header = [column]
        self.data = []

    def append(self, item):
        """Append data

        :param item: data
        """
        self.data.append(item)

    def success(self):
        """Append success event as green color
        """
        self.data.append([
            sys._getframe(1).f_code.co_name,
            TermColor.green('ok')])

    def fail(self):
        """Append fail event as red color
        """
        self.data.append([
            sys._getframe(1).f_code.co_name,
            TermColor.fail('fail')])

    def print_out(self):
        """Print out result
        """
        table = AsciiTable(self.header + self.data)
        print(table.table)


def clear_screen():
    """Clear screen
    """
    os.system('cls' if os.name == 'nt' else 'clear')


style = Style.from_dict({
    'completion-menu.completion': 'bg:#008888 #ffffff',
    'completion-menu.completion.current': 'bg:#00aaaa #000000',
    'scrollbar.background': 'bg:#88aaaa',
    'scrollbar.button': 'bg:#222222',
})


def get_full_path_of_props(cluster_id=-1, target='config'):
    """Get full path of props

    When user wants to edit props files, use it.

    :param cluster_id: cluster #
    :param target: target config (config | master | slave | thriftserver)
    """
    targets = {
        'config': 'config.yaml',
        'master': 'tsr2-conf/redis-master.conf.template',
        'slave': 'tsr2-conf/redis-slave.conf.template',
        'thriftserver': 'tsr2-conf/thriftserver.properties',
    }
    home = config.get_repo_cluster_path(cluster_id)
    f = targets[target]
    full_path = os.path.join(home, f)
    return full_path


def open_vim_editor(target='config'):
    """Open vim editor
    :param target: config | master | slave | thriftserver
    """
    cluster_id = config.get_cur_cluster_id()
    full_path = get_full_path_of_props(cluster_id, target)
    editor.edit(full_path)


def make_export_envs(ip, port):
    """Make export env
    """
    envs = config.get_env_dict(ip, port)
    cmd = '''\
export SR2_REDIS_HOME={sr2_redis_home} ; \
export SR2_REDIS_BIN={sr2_redis_bin} ; \
export SR2_REDIS_LIB={sr2_redis_lib} ; \
export SR2_REDIS_CONF={sr2_redis_conf} ; \
export SR2_REDIS_LOG={sr2_redis_log} ; \
export SR2_REDIS_DATA={sr2_redis_data} ; \
export SR2_REDIS_DUMP={sr2_redis_dump} ; \
export SR2_REDIS_DB_PATH={sr2_redis_db_path} ; \
export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:{ld_library_path} ; \
export DYLD_LIBRARY_PATH=$DYLD_LIBRARY_PATH:{dyld_library_path}''' \
        .format(**envs)
    return cmd


def convert_list_2_hyphen(ports):
    '''
    converted as shown below
    [1, 2, 3, 5, 7, 8, 10]
    =>
    ['1-3', '5', '7-8', 10]
    '''
    logger.debug('ports: {}'.format(ports))
    ret = []
    s = ports[0]
    pre = ports[0] - 1
    try:
        for port in ports:
            if pre != port - 1:
                if s != pre:
                    ret.append('{}-{}'.format(s, pre))
                else:
                    ret.append(str(s))
                s = port
            pre = port
        if s != pre:
            ret.append('{}-{}'.format(s, pre))
        else:
            ret.append(str(s))
        logger.debug('converted: {}'.format(ret))
        return ret
    except Exception:
        raise ConvertError("Invalid ports: '{}'".format(ports))


def is_number(target):
    try:
        target = target.decode('utf-8')
    except AttributeError:
        pass
    return target.isdecimal()


def to_str(target):
    if isinstance(target, bytes):
        target = target.decode('utf-8')
    return str(target)



import os

from ltcli import config
from ltcli import utils
from ltcli import net
from ltcli.log import logger
from ltcli.exceptions import PropsKeyError, FileNotExistError


CLEAN = 101
PENDING = 102
DEPLOYED = 103


class DeployUtil(object):

    def get_state(self, cluster_id, host='127.0.0.1'):
        path_of_fb = config.get_path_of_fb(cluster_id)
        cluster_path = path_of_fb['cluster_path']
        state_file = os.path.join(cluster_path, '.deploy.state')
        client = net.get_ssh(host)
        if net.is_exist(client, state_file):
            return PENDING
        if net.is_dir(client, cluster_path):
            return DEPLOYED
        return CLEAN

    def is_pending(self, cluster_id, nodes=['127.0.0.1']):
        if isinstance(nodes, str):
            nodes = [nodes]
        for node in nodes:
            path_of_fb = config.get_path_of_fb(cluster_id)
            cluster_path = path_of_fb['cluster_path']
            deploy_state = os.path.join(cluster_path, '.deploy.state')
            client = net.get_ssh(node)
            if net.is_exist(client, deploy_state):
                client.close()
                return True
            client.close()
        return False

    def transfer_installer(self, host, cluster_id, installer_path):
        installer_path = os.path.expanduser(installer_path)
        path_of_fb = config.get_path_of_fb(cluster_id)
        name = os.path.basename(installer_path)
        dst = os.path.join(path_of_fb['release_path'], name)

        client = net.get_ssh(host)
        sftp = net.get_sftp(client)

        if not net.is_dir(client, path_of_fb['release_path']):
            logger.debug("Not exist releases directory at '{}'".format(host))
            sftp.mkdir(path_of_fb['release_path'])
            logger.debug("Create releases directory at '{}'".format(host))

        logger.debug('Check {}...'.format(name))
        if not net.is_exist(client, dst):
            logger.debug('FAIL')
            logger.debug("Transfer '{}' to '{}'...".format(name, host))
            sftp.put(installer_path, '{}.download'.format(dst))
            command = 'mv {0}.download {0}'.format(dst)
            net.ssh_execute(client=client, command=command)
        sftp.close()
        client.close()
        logger.debug('OK')

    def install(self, host, cluster_id, name):
        logger.debug('Deploy cluster {} at {}...'.format(cluster_id, host))
        path_of_fb = config.get_path_of_fb(cluster_id)
        release_path = path_of_fb['release_path']
        cluster_path = path_of_fb['cluster_path']
        if '/' in name:
            name = os.path.basename(name)
        installer_path = os.path.join(release_path, name)
        command = '''chmod 755 {0}; \
            PATH=${{PATH}}:/usr/sbin; \
            {0} --full {1}'''.format(installer_path, cluster_path)
        client = net.get_ssh(host)
        if not net.is_exist(client, installer_path):
            raise FileNotExistError(installer_path, host=host)
        net.ssh_execute(client=client, command=command)
        client.close()
        logger.debug('OK')

    def get_meta_from_props(self, props_path):
        try:
            dic = config.get_props_as_dict(props_path)
            ret = [['hosts', '\n'.join(dic['sr2_redis_master_hosts'])]]
            value = dic['sr2_redis_master_ports']
            converted = utils.convert_list_2_hyphen(value)
            ret += [['master ports', ', '.join(converted)]]
            if config.is_key_enable(props_path, 'sr2_redis_slave_ports'):
                value = dic['sr2_redis_slave_ports']
                converted = utils.convert_list_2_hyphen(value)
                ret += [['slave ports', ', '.join(converted)]]
            ret += [
                ['ssd count', dic['ssd_count']],
                ['redis data path', dic['sr2_redis_data']],
                ['redis db path', dic['sr2_redis_db_path']],
                ['flash db path', dic['sr2_flash_db_path']],
            ]
            return ret
        except KeyError as key:
            raise PropsKeyError(str(key))

    def get_meta_from_dict(self, props_dict):
        meta = []
        meta.append(['hosts', '\n'.join(props_dict['hosts'])])
        converted = utils.convert_list_2_hyphen(props_dict['master_ports'])
        meta.append(['master ports', ', '.join(converted)])
        if props_dict['slave_ports']:
            converted = utils.convert_list_2_hyphen(props_dict['slave_ports'])
            meta.append(['slave ports', ', '.join(converted)])
        meta.append(['ssd count', props_dict['ssd_count']])
        meta.append(['db path', props_dict['prefix_of_db_path']])
        return meta



import logging
import re

import hiredis
import six
import sys
from retrying import retry
from six.moves import range

from ltcli.log import logger
from .clusternode import ClusterNode, base_balance_plan
from .connection import (
    CMD_CLUSTER_INFO,
    CMD_CLUSTER_NODES,
    CMD_INFO,
    Connection
)

SLOT_COUNT = 16384
PAT_CLUSTER_ENABLED = re.compile('cluster_enabled:([01])')
PAT_CLUSTER_STATE = re.compile('cluster_state:([a-z]+)')
PAT_CLUSTER_SLOT_ASSIGNED = re.compile('cluster_slots_assigned:([0-9]+)')
PAT_MIGRATING_IN = re.compile(r'\[([0-9]+)-<-(\w+)\]')
PAT_MIGRATING_OUT = re.compile(r'\[([0-9]+)->-(\w+)\]')


def _valid_node_info(n):
    return len(n) != 0 and 'handshake' not in n


def _ensure_cluster_status_unset(t):
    m = t.send_raw(CMD_INFO)
    logging.debug('Ask `info` Rsp %s', m)
    cluster_enabled = PAT_CLUSTER_ENABLED.findall(m)
    if len(cluster_enabled) == 0 or int(cluster_enabled[0]) == 0:
        raise hiredis.ProtocolError(
            'Node %s:%d is not cluster enabled' % (t.host, t.port))

    m = t.send_raw(CMD_CLUSTER_INFO)
    logging.debug('Ask `cluster info` Rsp %s', m)
    cluster_state = PAT_CLUSTER_STATE.findall(m)
    cluster_slot_assigned = PAT_CLUSTER_SLOT_ASSIGNED.findall(m)
    if cluster_state[0] != 'fail' or int(cluster_slot_assigned[0]) != 0:
        raise hiredis.ProtocolError(
            'Node %s:%d is already in a cluster' % (t.host, t.port))


def _ensure_cluster_status_set(t):
    m = t.send_raw(CMD_INFO)
    logging.debug('Ask `info` Rsp %s', m)
    cluster_enabled = PAT_CLUSTER_ENABLED.findall(m)
    if len(cluster_enabled) == 0 or int(cluster_enabled[0]) == 0:
        raise hiredis.ProtocolError(
            'Node %s:%d is not cluster enabled' % (t.host, t.port))

    m = t.send_raw(CMD_CLUSTER_INFO)
    logging.debug('Ask `cluster info` Rsp %s', m)
    cluster_state = PAT_CLUSTER_STATE.findall(m)
    cluster_slot_assigned = PAT_CLUSTER_SLOT_ASSIGNED.findall(m)
    if cluster_state[0] != 'ok' and int(cluster_slot_assigned[0]) == 0:
        raise hiredis.ProtocolError(
            'Node %s:%d is not in a cluster' % (t.host, t.port))


# Redis instance responses to clients BEFORE changing its 'cluster_state'
#   just retry some times, it should become OK
@retry(stop_max_attempt_number=128, wait_fixed=1000)
def _poll_check_status(t):
    m = t.send_raw(CMD_CLUSTER_INFO)
    logging.debug('Ask `cluster info` Rsp %s', m)
    cluster_state = PAT_CLUSTER_STATE.findall(m)
    cluster_slot_assigned = PAT_CLUSTER_SLOT_ASSIGNED.findall(m)
    if cluster_state[0] != 'ok' or int(cluster_slot_assigned[0]) != SLOT_COUNT:
        t.raise_('Unexpected status: %s' % m)


def _add_slots(conn, slots_list, max_slots):
    def addslots(slots_chunk):
        m = conn.execute('cluster', 'addslots', *slots_chunk)
        logging.debug('Ask `cluster addslots` Rsp %s', m)
        if m.lower() != 'ok':
            conn.raise_('Unexpected reply after ADDSLOTS: %s' % m)

    # split list to evenly sized chunks
    for i in range(0, len(slots_list), max_slots):
        addslots(slots_list[i:i + max_slots])


def _add_slots_range(conn, begin, end, max_slots):
    _add_slots(conn, list(range(begin, end)), max_slots)


def create(host_port_list, max_slots=1024):
    conns = []
    try:
        for host, port in set(host_port_list):
            t = Connection(host, port)
            conns.append(t)
            _ensure_cluster_status_unset(t)
            logging.info('Instance at %s:%d checked', t.host, t.port)

        logger.info('Cluster meet...')
        logger.info(' - {}:{}'.format(conns[0].host, conns[0].port))
        first_conn = conns[0]
        for i, t in enumerate(conns[1:]):
            logger.info(' - {}:{}'.format(t.host, t.port))
            t.execute('cluster', 'meet', first_conn.host, first_conn.port)

        slots_each = SLOT_COUNT // len(conns)
        slots_residue = SLOT_COUNT - slots_each * len(conns)
        first_node_slots = slots_residue + slots_each

        logger.info('Adding slots...')
        logger.info(' - {}:{}, {}'.format(
            first_conn.host,
            first_conn.port,
            slots_residue + slots_each,
        ))
        _add_slots_range(first_conn, 0, first_node_slots, max_slots)
        logging.info('Add %d slots to %s:%d', slots_residue + slots_each,
                     first_conn.host, first_conn.port)
        for i, t in enumerate(conns[1:]):
            msg = ' - {}:{}, {}'.format(t.host, t.port, slots_each)
            logger.info(msg)
            _add_slots_range(t, i * slots_each + first_node_slots,
                             (i + 1) * slots_each + first_node_slots,
                             max_slots)
            logging.info('Add %d slots to %s:%d', slots_each, t.host, t.port)
        logger.info('Check cluster state and asign slot...')
        for t in conns:
            _poll_check_status(t)
        logger.info('Ok')
    finally:
        for t in conns:
            t.close()


def start_cluster(host, port, max_slots=SLOT_COUNT):
    with Connection(host, port) as t:
        _ensure_cluster_status_unset(t)
        _add_slots_range(t, 0, SLOT_COUNT, max_slots)
        _poll_check_status(t)
        logging.info('Instance at %s:%d started as a standalone cluster', host,
                     port)


def start_cluster_on_multi(host_port_list, max_slots=SLOT_COUNT):
    return create(host_port_list, max_slots)


def _migr_keys(src_conn, target_host, target_port, slot):
    key_count = 0
    while True:
        keys = src_conn.execute('cluster', 'getkeysinslot', slot, 10)
        if len(keys) == 0:
            return key_count
        key_count += len(keys)
        src_conn.execute_bulk(
            [['migrate', target_host, target_port, k, 0, 30000] for k in keys])


def _migr_slots(source_node, target_node, slots, nodes):
    logging.info('Migrating %d slots from %s<%s:%d> to %s<%s:%d>', len(slots),
                 source_node.node_id, source_node.host, source_node.port,
                 target_node.node_id, target_node.host, target_node.port)
    key_count = 0
    for slot in slots:
        key_count += _migr_one_slot(source_node, target_node, slot, nodes)
    logging.info('Migrated: %d slots %d keys from %s<%s:%d> to %s<%s:%d>',
                 len(slots), key_count, source_node.node_id, source_node.host,
                 source_node.port, target_node.node_id, target_node.host,
                 target_node.port)


def _migr_one_slot(source_node, target_node, slot, nodes):
    def expect_exec_ok(m, conn, slot):
        if m.lower() != 'ok':
            conn.raise_('\n'.join([
                'Error while moving slot [ %d ] between' % slot,
                'Source node - %s:%d' % (source_node.host, source_node.port),
                'Target node - %s:%d' % (target_node.host, target_node.port),
                'Got %s' % m
            ]))

    @retry(stop_max_attempt_number=16, wait_fixed=100)
    def setslot_stable(conn, slot, node_id):
        m = conn.execute('cluster', 'setslot', slot, 'node', node_id)
        expect_exec_ok(m, conn, slot)

    source_conn = source_node.get_conn()
    target_conn = target_node.get_conn()

    try:
        expect_exec_ok(
            target_conn.execute('cluster', 'setslot', slot, 'importing',
                                source_node.node_id), target_conn, slot)
    except hiredis.ReplyError as e:
        if 'already the owner of' not in str(e):
            target_conn.raise_(str(e))

    try:
        expect_exec_ok(
            source_conn.execute('cluster', 'setslot', slot, 'migrating',
                                target_node.node_id), source_conn, slot)
    except hiredis.ReplyError as e:
        if 'not the owner of' not in str(e):
            source_conn.raise_(str(e))

    keys = _migr_keys(source_conn, target_node.host, target_node.port, slot)
    setslot_stable(source_conn, slot, target_node.node_id)
    for node in nodes:
        if node.master:
            setslot_stable(node.get_conn(), slot, target_node.node_id)
    sys.stdout.write('#')
    sys.stdout.flush()
    return keys


@retry(stop_max_attempt_number=8, wait_fixed=500)
def _meet(clst, new):
    m = clst.execute('cluster', 'meet', new.host, new.port)
    logging.debug('Ask `cluster meet` Rsp %s', m)
    if m.lower() != 'ok':
        clst.raise_('Unexpected reply after MEET: %s' % m)


def _join_to_cluster(clst, new):
    _ensure_cluster_status_set(clst)
    _ensure_cluster_status_unset(new)

    _meet(clst, new)
    _poll_check_status(new)


def join_cluster(cluster_host,
                 cluster_port,
                 newin_host,
                 newin_port,
                 balancer=None,
                 balance_plan=base_balance_plan):
    with Connection(newin_host, newin_port) as t, \
            Connection(cluster_host, cluster_port) as cnode:
        _join_to_cluster(cnode, t)
        nodes = []
        try:
            logging.info(
                'Instance at %s:%d has joined %s:%d; now balancing slots',
                newin_host, newin_port, cluster_host, cluster_port)
            nodes = _list_nodes(t, default_host=newin_host)[0]
            for src, dst, count in balance_plan(nodes, balancer):
                _migr_slots(src, dst, src.assigned_slots[:count], nodes)
        finally:
            for n in nodes:
                n.close()


def add_node(cluster_host, cluster_port, newin_host, newin_port):
    with Connection(newin_host, newin_port) as t, \
            Connection(cluster_host, cluster_port) as c:
        _join_to_cluster(c, t)


def join_no_load(cluster_host, cluster_port, newin_host, newin_port):
    return add_node(cluster_host, cluster_port, newin_host, newin_port)


def _check_master_and_migrate_slots(nodes, myself):
    other_masters = []
    master_ids = set()
    for node in nodes:
        if node.master:
            other_masters.append(node)
        else:
            master_ids.add(node.master_id)
    if len(other_masters) == 0:
        raise ValueError('This is the last node')
    if myself.node_id in master_ids:
        raise ValueError('The master still has slaves')

    mig_slots_to_each = len(myself.assigned_slots) // len(other_masters)
    for node in other_masters[:-1]:
        _migr_slots(myself, node, myself.assigned_slots[:mig_slots_to_each],
                    nodes)
        del myself.assigned_slots[:mig_slots_to_each]
    node = other_masters[-1]
    _migr_slots(myself, node, myself.assigned_slots, nodes)


def del_node(host, port):
    myself = None
    nodes = []
    t = Connection(host, port)
    try:
        _ensure_cluster_status_set(t)
        nodes, myself = _list_nodes(t, filter_func=_filter_not_failed)
        nodes.remove(myself)
        if myself.master:
            _check_master_and_migrate_slots(nodes, myself)
        logging.info('Migrated for %s / Broadcast a `forget`', myself.node_id)
        for node in nodes:
            tk = node.get_conn()
            try:
                tk.execute('cluster', 'forget', myself.node_id)
            except hiredis.ReplyError as e:
                if 'Unknown node' not in str(e):
                    raise
        t.execute('cluster', 'reset')
    finally:
        t.close()
        if myself is not None:
            myself.close()
        for n in nodes:
            n.close()


def quit_cluster(host, port):
    return del_node(host, port)


def shutdown_cluster(host, port, ignore_failed=False):
    with Connection(host, port) as conn:
        _ensure_cluster_status_set(conn)
        if ignore_failed:
            nodes = _list_nodes(conn, filter_func=_filter_not_failed)[0]
        else:
            nodes = _list_nodes(conn)[0]

        if len(nodes) > 1:
            conn.raise_('More than 1 nodes in cluster.')
        try:
            m = conn.execute('cluster', 'reset')
        except hiredis.ReplyError as e:
            if 'containing keys' in str(e):
                conn.raise_('Redis still contains keys')
            raise
        logging.debug('Ask `cluster delslots` Rsp %s', m)


def fix_migrating(host, port):
    nodes = dict()
    mig_srcs = []
    mig_dsts = []
    t = Connection(host, port)
    try:
        m = t.send_raw(CMD_CLUSTER_NODES)
        logging.debug('Ask `cluster nodes` Rsp %s', m)
        for node_info in m.split('\n'):
            if not _valid_node_info(node_info):
                continue
            node = ClusterNode(*node_info.split(' '))
            node.host = node.host or host
            nodes[node.node_id] = node

            mig_dsts.extend([(node, {
                'slot': g[0],
                'id': g[1]
            }) for g in PAT_MIGRATING_IN.findall(node_info)])
            mig_srcs.extend([(node, {
                'slot': g[0],
                'id': g[1]
            }) for g in PAT_MIGRATING_OUT.findall(node_info)])

        for n, args in mig_dsts:
            node_id = args['id']
            if node_id not in nodes:
                logging.error(
                    'Fail to fix %s:%d <- (referenced from %s:%d)'
                    ' - node %s is missing', n.host, n.port, host, port,
                    node_id)
                continue
            _migr_one_slot(nodes[node_id], n, int(args['slot']),
                           six.itervalues(nodes))
        for n, args in mig_srcs:
            node_id = args['id']
            if node_id not in nodes:
                logging.error(
                    'Fail to fix %s:%d -> (referenced from %s:%d)'
                    ' - node %s is missing', n.host, n.port, host, port,
                    node_id)
                continue
            _migr_one_slot(n, nodes[node_id], int(args['slot']),
                           six.itervalues(nodes))
    finally:
        t.close()
        for n in six.itervalues(nodes):
            n.close()


@retry(stop_max_attempt_number=200, wait_fixed=1000)
def _check_slave(slave_host, slave_port, t):
    slave_addr = '%s:%d' % (slave_host, slave_port)
    # logging.info('check_slave: %s' % slave_addr)
    for line in t.execute('cluster', 'nodes').split('\n'):
        if slave_addr in line:
            if 'slave' in line:
                return
            t.raise_('%s not switched to a slave' % slave_addr)


def replicate(master_host, master_port, slave_host, slave_port):
    with Connection(slave_host, slave_port) as t, \
            Connection(master_host, master_port) as master_conn:
        _ensure_cluster_status_set(master_conn)
        myself = _list_nodes(master_conn)[1]
        myid = myself.node_id if myself.master else myself.master_id

        _join_to_cluster(master_conn, t)
        logging.info('Instance at %s:%d has joined %s:%d; now set replica',
                     slave_host, slave_port, master_host, master_port)

        _replicate(t, myid)
        _check_slave(slave_host, slave_port, master_conn)
        logging.info('Instance at %s:%d set as replica to %s', slave_host,
                     slave_port, myid)


@retry(stop_max_attempt_number=8, wait_fixed=500)
def _replicate(t, myid):
    m = t.execute('cluster', 'replicate', myid)
    logging.debug('Ask `cluster replicate` Rsp %s', m)
    if m.lower() != 'ok':
        t.raise_('Unexpected reply after REPCLIATE: %s' % m)


def _alive_master(node):
    return node.master and not node.fail


def _filter_master(node):
    return node.master


def _filter_not_failed(node):
    return not node.fail


def _filter_not_failed_master(node):
    return node.master and not node.fail


def _list_nodes(conn, default_host=None, filter_func=lambda node: True):
    m = conn.send_raw(CMD_CLUSTER_NODES)
    logging.debug('Ask `cluster nodes` Rsp %s', m)
    default_host = default_host or conn.host

    nodes = []
    myself = None
    for node_info in m.split('\n'):
        if not _valid_node_info(node_info):
            continue
        node = ClusterNode(*node_info.split(' '))
        if 'myself' in node_info:
            myself = node
            if myself.host == '':
                myself.host = default_host
        if filter_func(node):
            nodes.append(node)
    return nodes, myself


def _list_masters(conn, default_host=None):
    return _list_nodes(
        conn, default_host or conn.host, filter_func=_filter_master)


def list_nodes(host, port, default_host=None, filter_func=lambda node: True):
    with Connection(host, port) as t:
        return _list_nodes(t, default_host or host, filter_func)


def list_masters(host, port, default_host=None):
    with Connection(host, port) as t:
        return _list_masters(t, default_host or host)


def custom_migrate_slots(src, dst, slots):
    src_host = src.info['ip']
    src_port = src.info['port']
    dst_host = dst.info['ip']
    dst_port = dst.info['port']
    if src_host == dst_host and src_port == dst_port:
        raise ValueError('Same node')
    with Connection(src_host, src_port) as t:
        nodes, myself = _list_masters(t, src_host)

    slots = set(slots)
    logging.debug('Migrating %s', slots)
    if not slots.issubset(set(myself.assigned_slots)):
        raise ValueError('Not all slot held by %s:%d' % (src_host, src_port))

    try:
        for n in nodes:
            if n.host == dst_host and n.port == dst_port:
                return _migr_slots(myself, n, slots, nodes)
        raise ValueError('Two nodes are not in the same cluster')
    finally:
        for n in nodes:
            n.close()


def migrate_slots(src_host, src_port, dst_host, dst_port, slots):
    if src_host == dst_host and src_port == dst_port:
        raise ValueError('Same node')
    with Connection(src_host, src_port) as t:
        nodes, myself = _list_masters(t, src_host)

    slots = set(slots)
    logging.debug('Migrating %s', slots)
    if not slots.issubset(set(myself.assigned_slots)):
        raise ValueError('Not all slot held by %s:%d' % (src_host, src_port))

    try:
        for n in nodes:
            if n.host == dst_host and n.port == dst_port:
                return _migr_slots(myself, n, slots, nodes)
        raise ValueError('Two nodes are not in the same cluster')
    finally:
        for n in nodes:
            n.close()


def rescue_cluster(host, port, subst_host, subst_port, max_slots=1024):
    failed_slots = set(range(SLOT_COUNT))
    nodes = []
    conn_subst = Connection(subst_host, subst_port)
    try:
        _ensure_cluster_status_unset(conn_subst)
        node_info = conn_subst.send_raw(CMD_CLUSTER_NODES).strip()
        node_subst = ClusterNode(*node_info.split(' '))

        with Connection(host, port) as conn_existing:
            _ensure_cluster_status_set(conn_existing)
            nodes = _list_nodes(
                conn_existing, filter_func=_filter_not_failed_master)[0]

        for node in nodes:
            failed_slots -= set(node.assigned_slots)
        if len(failed_slots) == 0:
            logging.info('No need to rescue cluster at %s:%d', host, port)
            return

        m = conn_subst.execute('cluster', 'meet', host, port)
        logging.debug('Ask `cluster meet` Rsp %s', m)
        if m.lower() != 'ok':
            conn_subst.raise_('Unexpected reply after MEET: %s' % m)

        _add_slots(conn_subst, list(failed_slots), max_slots)
        for slot in failed_slots:
            for node in nodes:
                m = node.get_conn().execute('cluster', 'setslot', slot, 'node',
                                            node_subst.node_id)
                if m.lower() != 'ok':
                    conn_subst.raise_('Unexpected reply after SETSLOT: %s' % m)
        _poll_check_status(conn_subst)
        for node in nodes:
            _poll_check_status(node.get_conn())
        logging.info('Instance at %s:%d serves %d slots to rescue the cluster',
                     subst_host, subst_port, len(failed_slots))
    finally:
        conn_subst.close()
        for node in nodes:
            node.close()


def execute(host, port, master_only, slave_only, commands):
    with Connection(host, port) as c:
        filter_func = lambda n: True
        if master_only:
            filter_func = _filter_master
        elif slave_only:
            filter_func = lambda n: n.slave
        nodes = _list_nodes(c, filter_func=filter_func)[0]

        result = []
        for n in nodes:
            r = None
            exc = None
            try:
                r = n.get_conn().execute(*commands)
            except Exception as e:
                exc = e
            result.append({
                'node': n,
                'result': r,
                'exception': exc,
            })
        return result



from __future__ import print_function

import os
import fileinput
import random
import subprocess
from threading import Thread
import time

from ltcli import config
from ltcli import utils
from ltcli.log import logger


class RedisCliUtil(object):
    @staticmethod
    def to_list_of_dict(target_lines):
        """convert list to list of dict

        :param target_lines: list
        :return: list of dict
        """
        l = []
        for line in target_lines:
            _, value_str = line.split(':')
            value_strs = value_str.split(',')
            d = {}
            for value_str in value_strs:
                key, value = value_str.split('=')
                d[key] = value
            l.append(d)
        return l

    @staticmethod
    def command(
        sub_cmd,
        cluster_id=-1,
        mute=False,
        formatter=None,
        host=None,
        port=0
    ):
        """Send redis-cli command

        :param sub_cmd: sub command
        :param cluster_id: target cluster #
        :param mute: without stdout
        :param formatter: If set, call formatter with output string
        :param host: host
        :param port: port
        :return: output string
        """
        ip_list = config.get_node_ip_list(cluster_id)
        port_list = config.get_master_port_list(cluster_id)
        if host:
            ip_list = [host]
            port_list = [port]
        outs = RedisCliUtil.command_raw(sub_cmd, ip_list, port_list)
        if mute:
            return outs
        if formatter:
            formatter(outs)
        else:
            logger.info(outs)
        return outs

    @staticmethod
    def command_all(sub_cmd, cluster_id=-1, formatter=None):
        """Send redis-cli command to all

        :param sub_cmd: sub command
        :param cluster_id: target cluster #
        :param formatter: if set, call formatter with output string
        """
        if cluster_id < 0:
            cluster_id = config.get_cur_cluster_id()
        master_host_list = config.get_master_host_list(cluster_id)
        master_port_list = config.get_master_port_list(cluster_id)
        slave_host_list = config.get_slave_host_list(cluster_id)
        slave_port_list = config.get_slave_port_list()
        outs, meta = RedisCliUtil.command_raw_all(
            sub_cmd,
            master_host_list,
            master_port_list
        )
        logger.debug(outs)
        buf = meta[:]
        outs, meta = RedisCliUtil.command_raw_all(
            sub_cmd,
            slave_host_list,
            slave_port_list
        )
        logger.debug(outs)
        buf += meta[1:]
        if formatter:
            formatter(buf)
        else:
            logger.debug(outs)

    @staticmethod
    def command_raw(sub_cmd, ip_list, port_list):
        """Send redis-cli command raw

        :param sub_cmd: sub command
        :param ip_list: ip list
        :param port_list: port list
        :return: output string
        """
        logger.debug('command_raw')
        targets = utils.get_ip_port_tuple_list(ip_list, port_list)
        count = len(targets)
        index = random.randrange(0, count)
        target = targets[index]
        ip, port = target
        # logger.info('redis-cli connect to %s:%s' % (ip, port))
        outs = ''
        redis_cli = os.path.join(config.get_tsr2_home(), 'bin', 'redis-cli')
        env = utils.make_export_envs(ip, port)
        command = '{env}; {redis_cli} -h {ip} -p {port} {sub_cmd}'.format(
            env=env,
            redis_cli=redis_cli,
            ip=ip,
            port=port,
            sub_cmd=sub_cmd)
        try:
            stdout = subprocess.check_output(command, shell=True)
            stdout = stdout.decode('utf-8')
            outs += stdout
        except subprocess.CalledProcessError as ex:
            logger.debug('exception: %s' % str(ex))
        logger.debug('subprocess: %s' % command)
        return outs

    @staticmethod
    def command_raw_all(sub_cmd, ip_list, port_list):
        """Send redis-cli command raw to all

        :param sub_cmd: sub command
        :param ip_list: ip list
        :param port_list: port list
        :return: (output string, meta)
        """
        logger.debug('command_raw_all')
        targets = utils.get_ip_port_tuple_list(ip_list, port_list)
        outs = ''
        stdout = ''
        meta = [['addr', 'stdout']]
        for ip, port in targets:
            env = utils.make_export_envs(ip, port)
            ex_cmd = os.path.join(config.get_tsr2_home(), 'bin', 'redis-cli')
            command = '{env}; {ex_cmd} -c -h {ip} -p {port} {sub_cmd}'.format(
                env=env,
                ex_cmd=ex_cmd,
                ip=ip,
                port=port,
                sub_cmd=sub_cmd
            )
            logger.debug('subprocess: %s' % command)
            try:
                stdout = subprocess.check_output(command, shell=True)
                stdout = stdout.decode('utf-8')
                outs += stdout
                meta.append(['%s:%s' % (ip, port), stdout])
            except subprocess.CalledProcessError as ex:
                logger.debug('exception: %s' % str(ex))
        return outs, meta

    @staticmethod
    def command_all_async(sub_cmd):
        def _async_target_func(m_s, pre_cmd, host, port, sub_cmd, ret):
            try:
                command = '{} -h {} -p {} {}'.format(pre_cmd, host, port, sub_cmd)
                logger.debug(command)
                stdout = subprocess.check_output(command, shell=True)
                stdout = stdout.decode('utf-8').strip()
                ret.append((m_s, host, port, 'OK', stdout))
            except Exception as ex:
                stderr = str(ex)
                logger.debug(stderr)
                ret.append((m_s, host, port, 'FAIL', stderr))
        cluster_id = config.get_cur_cluster_id()
        master_host_list = config.get_master_host_list(cluster_id)
        master_port_list = config.get_master_port_list(cluster_id)
        slave_host_list = config.get_slave_host_list(cluster_id)
        slave_port_list = config.get_slave_port_list(cluster_id)
        path_of_fb = config.get_path_of_fb(cluster_id)
        sr2_redis_bin = path_of_fb['sr2_redis_bin']

        logger.debug('command_all_async')
        cluster_id = config.get_cur_cluster_id()
        lib_path = config.get_ld_library_path(cluster_id)
        env_cmd = [
            'export LD_LIBRARY_PATH={};'.format(lib_path['ld_library_path']),
            'export DYLD_LIBRARY_PATH={};'.format(
                lib_path['dyld_library_path']
            ),
        ]
        env = ' '.join(env_cmd)
        threads = []
        ret = []  # (m/s, host, port, result, message)
        pre_cmd = '{} {}/redis-cli -c'.format(env, sr2_redis_bin)
        for host in master_host_list:
            for port in master_port_list:
                t = Thread(
                    target=_async_target_func,
                    args=('Master', pre_cmd, host, port, sub_cmd, ret),
                )
                threads.append(t)
        for host in slave_host_list:
            for port in slave_port_list:
                t = Thread(
                    target=_async_target_func,
                    args=('Slave', pre_cmd, host, port, sub_cmd, ret),
                )
                threads.append(t)
        for th in threads:
            th.start()
            time.sleep(0.02)
        for th in threads:
            th.join()
        return ret

    @staticmethod
    def save_redis_template_config(key, value):
        """Save redis template config to file

        :param key: key
        :param value: value
        """
        key = key.strip()
        cluster_id = config.get_cur_cluster_id()
        path_of_fb = config.get_path_of_fb(cluster_id)
        master_template = path_of_fb['master_template']
        slave_template = path_of_fb['slave_template']
        RedisCliUtil._save_config(master_template, key, value)
        RedisCliUtil._save_config(slave_template, key, value)

    @staticmethod
    def _save_config(f, key, value):
        inplace_count = 0
        for line in fileinput.input(f, inplace=True):
            if line.startswith(key):
                msg = '{key} {value}'.format(key=key, value=value)
                inplace_count += 1
                print(msg)
            else:
                print(line, end='')
        logger.debug('inplace: %d (%s)' % (inplace_count, f))
        if inplace_count == 1:
            logger.info('save config(%s) success' % f)
        else:
            logger.warn('save config(%s) fail(%d)' % (f, inplace_count))



import re


def get_word_between(target, front_regex, back_regex):
    '''Return word between front regex and back regex in target
    if it is muliple, return only first
    if there is not, return empty string

    :param target: target string
    :param front_regex: regular expression
    :param back_regex: regular expression
    '''
    front_pattern = re.compile(front_regex)
    front_match = front_pattern.search(target)
    back_pattern = re.compile(back_regex)
    back_match = back_pattern.search(target)
    if not front_match or not back_match:
        return ''
    return str(target[front_match.end():back_match.start()])



import os
import sys

from logbook import Logger, Processor, StreamHandler, RotatingFileHandler
from logbook import DEBUG, INFO, WARNING, ERROR

from ltcli import color


def get_log_color(level):
    color_by_level = {
        DEBUG: color.ENDC,
        INFO: color.WHITE,
        WARNING: color.YELLOW,
        ERROR: color.RED,
    }
    return color_by_level[level]


def get_log_code(level):
    code = {
        'debug': DEBUG,
        'info': INFO,
        'warn': WARNING,
        'warning': WARNING,
        'error': ERROR,
    }
    return code[level]


formatter = {
    'screen': '' .join([
        '{record.extra[level_color]}',
        '{record.message}',
        '{record.extra[clear_color]}',
    ]),
    'screen_detail': ''.join([
        '{record.time:%Y-%m-%d %H:%M:%S}',
        ' ',
        '[',
        '{record.extra[level_color]}',
        '{record.level_name:<7}',
        '{record.extra[clear_color]}',
        ']',
        ' ',
        '({record.extra[basename]} --- {record.func_name}():{record.lineno})',
        ' : ',
        '{record.extra[level_color]}',
        '{record.message}',
        '{record.extra[clear_color]}',
    ]),
    'file': ''.join([
        '{record.time:%Y-%m-%d %H:%M:%S}',
        ' ',
        '[{record.level_name:<7}]',
        ' ',
        '({record.extra[basename]} --- {record.func_name}():{record.lineno})',
        ' : ',
        '{record.message}',
    ]),
}


def inject_extra(record):
    record.extra['basename'] = os.path.basename(record.filename)
    record.extra['level_color'] = get_log_color(record.level)
    record.extra['clear_color'] = color.ENDC


logger = Logger('root')

# extra info
processor = Processor(inject_extra)
processor.push_application()

# for screen log
screen_level = INFO
stream_handler = StreamHandler(sys.stdout, level=screen_level, bubble=True)
stream_handler.format_string = formatter['screen']
stream_handler.push_application()

# for rolling file log
p = os.environ['FBPATH']
if not os.path.isdir(p):
    os.system('mkdir -p {}'.format(p))
file_path = os.path.expanduser(os.path.join(p, 'logs'))
if os.path.isdir(file_path):
    backup_count = 7
    max_size = 1024 * 1024 * 1024  # 1Gi
    file_level = DEBUG
    each_size = max_size / (backup_count + 1)
    filename = os.path.join(file_path, 'ltcli-rotate.log')
    rotating_file_handler = RotatingFileHandler(
        filename=filename,
        level=file_level,
        bubble=True,
        max_size=each_size,
        backup_count=backup_count
    )
    rotating_file_handler.format_string = formatter['file']
    rotating_file_handler.push_application()
    logger.debug('start logging on file: {}'.format(filename))
else:
    try:
        os.mkdir(file_path)
    except Exception:
        logger.error("CreateDirError: {}".format(file_path))
        logger.warn('Could not logging in file. Confirm and restart.')


def set_level(level):
    level_list = ['debug', 'info', 'warning', 'error', 'warn']
    if level not in level_list:
        level_list.remove('warn')
        logger.error("LogLevelError: '{}'. Select in {}".format(
            level,
            level_list
        ))
        return
    code = get_log_code(level)
    stream_handler.level = code
    print(color.white('Changed log level to {}'.format(level)))


def set_mode(mode):
    mode_list = ['debug', 'normal']
    if mode not in mode_list:
        logger.error('ModeError: {}. Select in {}'.format(mode, mode_list))
        return
    if mode == 'debug':
        stream_handler.format_string = formatter['screen_detail']
        set_level('debug')
    if mode == 'normal':
        stream_handler.format_string = formatter['screen']
        set_level('info')



import os

from ltcli import color


necessary_env = []
export_example = {
    'FBPATH': 'export FBPATH=$HOME/.flashbase',
    'LANG': 'export LANG=en_US.utf-8',
}

if 'FBPATH' not in os.environ:
    necessary_env.append('FBPATH')
if 'LANG' not in os.environ:
    necessary_env.append('LANG')

if necessary_env:
    msg = [
        'To start using ltcli, you should set env {}'.format(
            ', '.join(necessary_env)
        ),
        'ex)',
    ]
    for env in necessary_env:
        msg.append(export_example[env])
    print(color.red('\n'.join(msg)))
    exit(1)

os.environ['LC_ALL'] = os.environ['LANG']



import subprocess as sp
import os
import re
import time

from ltcli import config
from ltcli.log import logger
from ltcli.exceptions import FileNotExistError, EnvError


# pylint: disable=line-too-long, anomalous-backslash-in-string
ROLLING_LOGFILE_REGEX = 'spark.*rolling.*org\.apache\.spark\.sql\.hive\.thriftserver\.HiveThriftServer2-1.*\.log'
NOHUP_LOGFILE_REGEX = 'spark.*org\.apache\.spark\.sql\.hive\.thriftserver\.HiveThriftServer2-1.*\.out'
ROLLING_LOGFILE = 'spark*rolling*org.apache.spark.sql.hive.thriftserver.HiveThriftServer2-1*.log'
NOHUP_LOGFILE = 'spark*org.apache.spark.sql.hive.thriftserver.HiveThriftServer2-1*.out'


def _get_env():
    if 'SPARK_HOME' not in os.environ:
        raise EnvError('you should set env SPARK_HOME')
    ret = {
        'spark_home': os.environ['SPARK_HOME'],
    }
    env_list = ['SPARK_BIN', 'SPARK_SBIN', 'SPARK_LOG', 'SPARK_CONF']
    cluster_id = config.get_cur_cluster_id()
    path_of_fb = config.get_path_of_fb(cluster_id)
    ths_props_path = path_of_fb['thrift_properties']
    for env in env_list:
        cmd = 'source {}; echo ${}'.format(ths_props_path, env.upper())
        stdout = sp.check_output(cmd, shell=True).decode('utf-8').strip()
        ret[env.lower()] = stdout
    return ret


def _check_spark():
    spark_env = _get_env()
    if not os.path.isdir(spark_env['spark_home']):
        raise FileNotExistError(spark_env['spark_home'])
    if not os.path.isdir(spark_env['spark_bin']):
        raise FileNotExistError(spark_env['spark_bin'])
    if not os.path.isdir(spark_env['spark_sbin']):
        raise FileNotExistError(spark_env['spark_sbin'])


def _get_hive_opts_str():
    ret = [
        '--hiveconf hive.server2.thrift.bind.host=$HIVE_HOST \\',
        '--hiveconf hive.server2.thrift.port=$HIVE_PORT \\',
        '--master yarn \\',
        '--executor-memory $EXECUTOR_MEMORY \\',
        '--driver-memory $DRIVER_MEMORY \\',
        '--num-executors $EXECUTERS \\',
        '--conf spark.ui.port=$SPARK_UI_PORT \\',
        '--conf spark.metrics.conf=$SPARK_METRICS \\',
        '--conf spark.executor.cores=$EXECUTER_CORES \\',
        '--conf spark.executor.extraClassPath=$EXECUTOR_CLASSPATH \\',
        '--driver-class-path $DRIVER_CLASSPATH \\',
        '--conf spark.eventLog.enabled=$EVENT_LOG_ENABLED \\',
        '--conf spark.eventLog.dir=$EVENT_LOG_DIR \\',
    ]
    return '\n'.join(ret)


def _find_files_with_regex(dir_path, pattern):
    ret = []
    for filename in os.listdir(dir_path):
        if re.search(pattern, filename):
            ret.append(filename)
    return ret


class ThriftServer(object):
    def beeline(self, **kargs):
        """Connect to thriftserver command line
        """
        logger.debug('thriftserver_command_beeline')
        _check_spark()
        cluster_id = config.get_cur_cluster_id()
        path_of_fb = config.get_path_of_fb(cluster_id)
        ths_props_path = path_of_fb['thrift_properties']
        cmd = 'source {}; echo ${}'.format(ths_props_path, 'HIVE_HOST')
        host = sp.check_output(cmd, shell=True).decode('utf-8').strip()
        cmd = 'source {}; echo ${}'.format(ths_props_path, 'HIVE_PORT')
        port = sp.check_output(cmd, shell=True).decode('utf-8').strip()
        spark_env = _get_env()
        base_cmd = '{}/beeline'.format(spark_env['spark_bin'])
        options = {
            'u': 'jdbc:hive2://{}:{}'.format(host, port),
            'n': os.environ['USER']
        }
        for key, value in kargs.items():
            options[key] = value
        for key, value in options.items():
            base_cmd += ' -{} {}'.format(key, value)
        logger.debug(base_cmd)
        logger.info('Connecting...')
        os.system(base_cmd)

    def start(self):
        """Start thriftserver
        """
        logger.debug('thriftserver_command_start')
        _check_spark()
        cluster_id = config.get_cur_cluster_id()
        path_of_fb = config.get_path_of_fb(cluster_id)
        ths_props_path = path_of_fb['thrift_properties']
        source_cmd = 'source {}'.format(ths_props_path)
        hive_opts = _get_hive_opts_str()
        base_cmd = '$SPARK_SBIN/start-thriftserver.sh {}'.format(hive_opts)
        cmd = '{}; {}'.format(source_cmd, base_cmd)
        logger.debug(cmd)
        os.system(cmd)
        spark_log = os.path.join(os.environ['SPARK_HOME'], 'logs')
        if _find_files_with_regex(spark_log, ROLLING_LOGFILE_REGEX):
            for file in _find_files_with_regex(spark_log, NOHUP_LOGFILE_REGEX):
                os.remove(os.path.join(spark_log, file))

    def stop(self):
        """Stop thriftserver
        """
        logger.debug('thriftserver_command_stop')
        _check_spark()
        cluster_id = config.get_cur_cluster_id()
        path_of_fb = config.get_path_of_fb(cluster_id)
        ths_props_path = path_of_fb['thrift_properties']
        source_cmd = 'source {}'.format(ths_props_path)
        base_cmd = '$SPARK_SBIN/stop-thriftserver.sh'
        cmd = '{}; {}'.format(source_cmd, base_cmd)
        logger.debug(cmd)
        os.system(cmd)

    def monitor(self, n=10):
        """Show thriftserver log
        """
        logger.debug('thriftserver_command_monitor')
        _check_spark()
        cluster_id = config.get_cur_cluster_id()
        path_of_fb = config.get_path_of_fb(cluster_id)
        ths_props_path = path_of_fb['thrift_properties']
        source_cmd = 'source {}'.format(ths_props_path)
        spark_log = _get_env()['spark_log']
        # log_file_path = ''.join([
        #     spark_log,
        #     '/spark-{}-org.apache'.format(os.environ['USER']),
        #     '.spark.sql.hive.thriftserver.HiveThriftServer2-1-*.out',
        # ])
        log_file_path = os.path.join(spark_log, NOHUP_LOGFILE)
        if _find_files_with_regex(spark_log, ROLLING_LOGFILE_REGEX):
            log_file_path = os.path.join(spark_log, ROLLING_LOGFILE)
        base_cmd = 'tail -F -n {} {}'.format(n, log_file_path)
        cmd = '{}; {}'.format(source_cmd, base_cmd)
        logger.debug(cmd)
        logger.info('Press Ctrl-C for exit.')
        os.system(cmd)

    def restart(self):
        """Thriftserver restart
        """
        self.stop()
        time.sleep(2)
        self.start()



import socket
from functools import wraps

import hiredis
import six
from six import b

from .exceptions import RedisIOError, RedisStatusError

SYM_STAR = b('*')
SYM_DOLLAR = b('$')
SYM_CRLF = b('\r\n')
EMPTY = b('')

ENCODING = 'utf-8'


def encode(value):
    if isinstance(value, six.binary_type):
        return value
    if isinstance(value, six.integer_types):
        return b(str(value))
    if isinstance(value, float):
        return b(repr(value))
    if isinstance(value, six.text_type):
        return value.encode(ENCODING)
    if not isinstance(value, six.string_types):
        return b(value)
    return value


def squash_commands(commands):
    output = []
    buf = EMPTY

    for c in commands:
        buf = EMPTY.join((buf, SYM_STAR, b(str(len(c))), SYM_CRLF))

        for arg in map(encode, c):
            if len(buf) > 6000 or len(arg) > 6000:
                output.append(
                    EMPTY.join((buf, SYM_DOLLAR, b(str(len(arg))), SYM_CRLF)))
                output.append(arg)
                buf = SYM_CRLF
            else:
                buf = EMPTY.join((buf, SYM_DOLLAR, b(str(len(arg))), SYM_CRLF,
                                  arg, SYM_CRLF))
    output.append(buf)
    return output


def pack_command(command, *args):
    return squash_commands([(command,) + args])


CMD_INFO = pack_command('info')
CMD_CLUSTER_NODES = pack_command('cluster', 'nodes')
CMD_CLUSTER_INFO = pack_command('cluster', 'info')


def _wrap_sock_op(f):
    @wraps(f)
    def g(conn, *args, **kwargs):
        try:
            return f(conn, *args, **kwargs)
        except IOError as e:
            raise RedisIOError(e, conn.host, conn.port)

    return g


class Connection(object):
    def __init__(self, host, port, timeout=5):
        self.host = host
        self.port = port
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.reader = hiredis.Reader()
        self.last_raw_message = EMPTY

        self.sock.settimeout(timeout)
        self._conn()

    @_wrap_sock_op
    def _conn(self):
        self.sock.connect((self.host, self.port))

    @_wrap_sock_op
    def _recv(self):
        while True:
            m = self.sock.recv(16384)
            self.last_raw_message += m
            self.reader.feed(m)
            r = self.reader.gets()
            # From hiredis.Reader : https://github.com/redis/hiredis-py#usage
            # > When the buffer does not contain a full reply, gets returns False.
            if r != False:
                return r

    @_wrap_sock_op
    def _recv_multi(self, n):
        resp = []
        while len(resp) < n:
            m = self.sock.recv(16384)
            self.last_raw_message += m
            self.reader.feed(m)

            r = self.reader.gets()
            # See the previous comment
            while r != False:
                resp.append(r)
                r = self.reader.gets()
        return resp

    @_wrap_sock_op
    def send_raw(self, command, recv=None):
        recv = recv or self._recv
        for c in command:
            self.sock.send(c)
        r = recv()
        if r is None:
            raise ValueError('No reply')
        if isinstance(r, hiredis.ReplyError):
            raise r

        if isinstance(r, list):
            return [i.decode(ENCODING) for i in r]
        return r.decode(ENCODING)

    def execute(self, *args):
        return self.send_raw(pack_command(*args))

    def execute_bulk(self, cmd_list):
        return self.send_raw(
            squash_commands(cmd_list),
            recv=lambda: self._recv_multi(len(cmd_list)))

    def close(self):
        return self.sock.close()

    def raise_(self, message):
        raise RedisStatusError(message, self.host, self.port)

    def __enter__(self):
        return self

    def __exit__(self, except_type, except_obj, tb):
        self.close()
        return False

    def talk_raw(self, command, recv=None):
        return self.send_raw(command, recv)

    def talk(self, *args):
        return self.execute(*args)

    def talk_bulk(self, cmd_list):
        return self.execute_bulk(cmd_list)



import os
import subprocess

EDITOR = os.environ.get('EDITOR', 'vim')


def edit(f, syntax=None):
    """Open editor

    :param f: file name
    """
    cmd = [EDITOR]
    if syntax and EDITOR == 'vim':
        cmd.append('-c "set syntax={}"'.format(syntax))
    cmd.append(f)
    cmd = ' '.join(cmd)
    subprocess.call(cmd, shell=True)



from ask import ask, askBool, askInt

from .command import custom_migrate_slots

msg_unknown_or_master = '*** The specified node is not known or not a master, ' \
                        '' \
                        '' \
                        'please retry.'


class CustomReshard(object):
    def __init__(self, trib, slots, src_node_ids, dest_node_id, src_all,
        config):
        self.trib = trib
        self.move_slot_count = slots
        self.src_node_ids = src_node_ids
        self.dest_node_id = dest_node_id
        self.src_all = src_all
        self.config = config

    def run(self):
        while self.move_slot_count <= 0:
            self._set_slots()
        while self.dest_node_id is None:
            self._set_dest_node_id()
        if not self.src_all and len(self.src_node_ids) == 0:
            self._set_src_node_ids()
        table = self._ready_plan()
        self._show_plan(table)
        self._execute_plan(table)

    def _ready_plan(self):
        print('ready_plan')
        reshard_table = self.trib.compute_reshard_table(
            self.src_node_ids,
            self.move_slot_count)
        return reshard_table

    def _show_plan(self, table):
        print('  Ready to move %s slots' % self.move_slot_count)
        self._print_src_nodes()
        self._print_dest_node()
        print('')
        for item in table:
            node = item['node']
            name = node.info['name']
            slot_num_list = item['slot_num_list']
            for slot_num in slot_num_list:
                print('Moving slot %s from %s' % (slot_num, name))
        print('')

    def _print_src_nodes(self):
        print('  Source nodes:')
        for src_node_id in self.src_node_ids:
            print('    %s' % self.trib.get_node_by_name(
                src_node_id).info_string())

    def _print_dest_node(self):
        print('  Destination node:')
        print('    %s' % self.trib.get_node_by_name(
            self.dest_node_id).info_string())

    def _execute_plan(self, table):
        text = 'Do you want to proceed with the proposed reshard plan?'
        t = self.trib
        if askBool(text):
            dest = t.name_to_node(self.dest_node_id)
            for item in table:
                node = item['node']
                slot_num_list = item['slot_num_list']
                src = node
                slots = slot_num_list
                custom_migrate_slots(src, dest, slots)
        else:
            print('Reshard canceled')

    def _set_slots(self):
        total_slot_count = self.config['total_slot_count']
        text = 'How many slots do you want to move (from 1 to %d)?' % \
               total_slot_count
        self.move_slot_count = int(askInt(text))

    def _set_dest_node_id(self):
        text = 'What is the receiving node ID?'
        candidate = ask(text)
        if self.trib.is_exist_master_node_id(candidate):
            self.dest_node_id = candidate
        else:
            print(msg_unknown_or_master)

    def _set_src_node_ids(self):
        text = '''Please enter all the source node IDs.
  Type 'all' to use all the nodes as source nodes for the hash slots.
  Type 'done' once you entered all the source nodes IDs.'''
        while True:
            print('src node ids:', self.src_node_ids)
            candidate = ask(text)
            if candidate == 'all':
                self.src_all = True
                tmp = filter(
                    lambda x: self.dest_node_id != x.info['name'],
                    self.trib.master_nodes)
                self.src_node_ids = list(map(lambda x: x.info['name'], tmp))
                return
            elif candidate == 'done':
                return
            if self.dest_node_id == candidate or candidate in self.src_node_ids:
                print(msg_unknown_or_master)
                continue
            if self.trib.is_exist_master_node_id(candidate):
                self.src_node_ids.append(candidate)
            else:
                print(msg_unknown_or_master)



from __future__ import print_function

from terminaltables import AsciiTable

from ltcli import config
from ltcli import utils
from ltcli import color
from ltcli.center import Center
from ltcli.rediscli_util import RedisCliUtil
from ltcli.utils import TableReport
from ltcli.log import logger


class RedisCliInfo(object):
    def __init__(self):
        pass

    def all(self, host=None, port=None):
        """Command: cli info all"""
        RedisCliUtil.command(sub_cmd='info', host=host, port=port)

    def memory(self, host=None, port=None):
        """Command: cli info memory"""
        RedisCliUtil.command(sub_cmd='info memory', host=host, port=port)

    def eviction(self, host=None, port=None):
        """Command: cli info eviction"""
        RedisCliUtil.command(sub_cmd='info eviction', host=host, port=port)

    def keyspace(self, host=None, port=None):
        """Command: cli info keyspace"""
        RedisCliUtil.command(sub_cmd='info keyspace', host=host, port=port)

    def tablespace(self, host=None, port=None):
        """Command: cli info tablespace"""
        RedisCliUtil.command(sub_cmd='info tablespace', host=host, port=port)

    def replication(self, host=None, port=None):
        """Command: cli info replication"""
        RedisCliUtil.command('info replication', host=host, port=port)


class RedisCliCluster(object):
    def __init__(self):
        pass

    def info(self):
        """Command: cli cluster info"""
        RedisCliUtil.command('cluster info')

    def nodes(self):
        """Command: cli cluster nodes"""
        RedisCliUtil.command('cluster nodes')

    def slots(self):
        """Command: cli cluster slots"""
        def formatter(outs):
            lines = outs.splitlines()
            replicas = config.get_replicas()
            column_count = 2 + 2 * (replicas + 1)
            row_count = len(lines) / column_count
            header = [['start', 'end', 'm_ip', 'm_port']]
            remain = column_count - 4
            data = []
            for i in range(0, remain / 2):
                header[0].append('s_ip_%d' % i)
                header[0].append('s_port_%d' % i)
            for i in range(0, row_count):
                data.append(lines[i * column_count: column_count * (i + 1)])
            data.sort(key=lambda x: int(x[0]))
            table = AsciiTable(header + data)
            print(table.table)

        RedisCliUtil.command('cluster slots', formatter=formatter)


class RedisCliConfig(object):
    def __init__(self):
        pass

    def get(self, key, all=False, host=None, port=None):
        """Command: cli config get [key]

        :param key: redis config keyword
        :param all: If True, command to all nodes
        :param host: host
        :param port: port
        """
        if not isinstance(all, bool):
            logger.error("option '--all' can use only 'True' or 'False'")
            return
        if (not host or not port) and not all:
            logger.error("Enter host and port or use '--all' option.")
            return
        sub_cmd = 'config get "{key}" 2>&1'.format(key=key)
        if all:
            meta = []
            ret = RedisCliUtil.command_all_async(sub_cmd)
            for m_s, host, port, result, message in ret:
                addr = '{}:{}'.format(host, port)
                if result == 'OK':
                    _, value = message.split('\n')
                    meta.append([m_s, addr, value])
                else:
                    meta.append([m_s, addr, color.red(result)])
            utils.print_table([['TYPE', 'ADDR', 'RESULT']] + meta)
            return
        if all:
            RedisCliUtil.command_all(
                sub_cmd=sub_cmd,
                formatter=utils.print_table)
        else:
            RedisCliUtil.command(
                sub_cmd=sub_cmd,
                host=host,
                port=port)

    def set(self, key, value, all=False, save=False, host=None, port=None):
        """Command: cli config set [key] [value]

        :param key: redis config keyword
        :param value: value
        :param save: If True, save value to config file
        :param all: If True, command to all nodes
        :param host: host
        :param port: port
        """
        if not isinstance(all, bool):
            logger.error("option '--all' can use only 'True' or 'False'")
            return
        if not isinstance(save, bool):
            logger.error("option '--save' can use only 'True' or 'False'")
            return
        if (not host or not port) and not all:
            logger.error("Enter host and port or use '--all' option.")
            return
        sub_cmd = 'config set {key} {value} 2>&1'.format(key=key, value=value)
        if all:
            meta = []
            ret = RedisCliUtil.command_all_async(sub_cmd)
            ok_cnt = 0
            for m_s, host, port, result, _ in ret:
                addr = '{}:{}'.format(host, port)
                if result == 'OK':
                    ok_cnt += 1
                else:
                    meta.append([m_s, addr, color.red('FAIL')])
            if meta:
                utils.print_table([['TYPE', 'ADDR', 'RESULT']] + meta)
            logger.info('success {}/{}'.format(ok_cnt, len(ret)))
        else:
            RedisCliUtil.command(
                sub_cmd=sub_cmd,
                host=host,
                port=port)
        if save:
            RedisCliUtil.save_redis_template_config(key, value)
            center = Center()
            center.update_ip_port()
            success = center.check_hosts_connection()
            if not success:
                return
            center.configure_redis()
            center.sync_conf()



import os
import re

# pylint: disable=unused-import
# using like 'ask_util.askBool'
from ask import ask, askInt, askBool

from ltcli.log import logger
from ltcli import config
from ltcli import net
from ltcli import utils


START_PORT = 18000
MASTER_OFFSET = 100
SLAVE_OFFSET = 50
PORT_MININUM = 18000
PORT_MAXIMUM = 65535


def hosts(save, default=None):
    logger.debug('ask host')
    deploy_history = config.get_deploy_history()
    if not default:
        default = deploy_history['hosts']
    q = 'Please type host list separated by comma(,)'
    result = ask(q, default=', '.join(default))
    result = list(map(lambda x: x.strip(), result.split(',')))
    if save:
        deploy_history['hosts'] = result
        config.save_deploy_history(deploy_history)
    logger.info('OK, {}'.format(result))
    return result


def installer():
    '''
    Select installer from list of '$FBPATH/releases'
    or input absolute path of file directly
    return installer path
    '''
    logger.debug('ask installer')
    path_of_cli = config.get_path_of_cli(None)
    release_path = path_of_cli['release_path']
    if not os.path.exists(release_path):
        os.mkdir(release_path)
    installer_list = net.get_installers_from_fb_s3()
    buf = os.listdir(release_path)
    buf = list(filter(lambda x: x != '.gitignore', buf))
    pattern = '.download'
    buf = list(filter(lambda x: pattern not in x, buf))
    for file_name in buf:
        installer_list.append({
            'name': file_name,
            'url': os.path.join(release_path, file_name),
            'type': 'local',
        })

    # formatting msg
    formatted = []
    for i, obj in enumerate(installer_list):
        formatted.append('    ({index}) [{type}] {name}'.format(
            index=i+1,
            name=obj['name'],
            type=obj['type'].upper(),
        ))
    msg = [
        'Select installer',
        '',
        '    [ INSTALLER LIST ]',
        '{}\n'.format('\n'.join(formatted)),
        'Please enter the number, file path or url of the installer you want to use.',
        "you can also add file in list by copy to '$FBPATH/releases/'",
    ]
    if not installer_list:
        msg = [
            'Select installer',
            '',
            '    [ INSTALLER LIST ]',
            '    (empty)\n\n'
            'Please enter file path or url of the installer you want to use',
            "you can also add file in list by copy to '$FBPATH/releases/'",
        ]

    result = ask('\n'.join(msg))
    while True:
        result = result.strip()
        if installer_list and utils.is_number(result):
            # case: select in list
            result = int(result) - 1
            if result in range(0, len(installer_list)):
                selected = installer_list[result]
                if selected['type'] == 'download':
                    url = selected['url']
                    file_name = selected['name']
                    installer_path = os.path.join(release_path, file_name)
                    success = net.download_file(url, installer_path)
                    if success:
                        logger.info('OK, {}'.format(file_name))
                        return installer_path
                    msg = [
                        "Fail to download from '{}', ".format(url),
                        'try again.'
                    ]
                    logger.error(''.join(msg))
                if selected['type'] == 'local':
                    ret = selected['url']
                    logger.debug('Select insaller in list: {}'.format(ret))
                    logger.info('OK, {}'.format(selected['name']))
                    return os.path.expanduser(ret)
            msg = [
                'Choose a number ',
                'between 1 and {}'.format(len(installer_list)),
                ', please try again'
            ]
            logger.error(''.join(msg))
        elif result.startswith(('~', '/')):
            # case: type path
            if os.path.isfile(os.path.expanduser(result)):
                logger.debug('Select insaller by path: {}'.format(result))
                logger.info('OK, {}'.format(os.path.basename(result)))
                return os.path.expanduser(result)
            msg = [
                "File not existed: '{}'".format(result),
                ', please try again'
            ]
            logger.error(''.join(msg))
        elif result.startswith(('http://', 'https://')):
            # case: type url
            url = result
            file_name = url.split('?')[0].split('/')[-1]
            installer_path = os.path.join(release_path, file_name)
            success = net.download_file(url, installer_path)
            if success:
                logger.info('OK, {}'.format(file_name))
                return installer_path
            logger.error("Fail to download from '{}', try again".format(url))
        else:
            msg = [
                "Invalid input: '{}', ".format(result),
                "please try again",
            ]
            logger.error(''.join(msg))
        result = ask('')


def port_range_safe(port):
    if port < PORT_MININUM:
        return False
    if port > PORT_MAXIMUM:
        return False
    return True


def master_ports(save, cluster_id, default_count=None):
    logger.debug('ask master ports')
    deploy_history = config.get_deploy_history()
    if not default_count:
        default_count = deploy_history['master_count']
    q = 'How many masters would you like to create on each host?'
    m_count = int(askInt(q, default=str(default_count)))
    if m_count <= 0:
        logger.warn("The number of master must be greater than 0. try again.")
        return master_ports(cluster_id, default_count)
    logger.info('OK, {}'.format(m_count))
    if save:
        deploy_history['master_count'] = m_count
        config.save_deploy_history(deploy_history)

    start_m_ports = START_PORT + cluster_id * MASTER_OFFSET
    end_m_ports = start_m_ports + m_count - 1
    if start_m_ports == end_m_ports:
        default_m_ports = str(start_m_ports)
    else:
        default_m_ports = '{}-{}'.format(start_m_ports, end_m_ports)

    q = [
        'Please type ports separate with comma(,) ',
        'and use hyphen(-) for range.',
    ]
    while True:
        result = ask(''.join(q), default=default_m_ports)
        result = list(map(lambda x: x.strip(), result.split(',')))
        valid = True
        m_ports = set()
        pattern = re.compile('[0-9]+-[0-9]+')
        for item in result:
            # range number
            matched = pattern.match(item)
            if matched:
                s, e = map(int, item.split('-'))
                if s > e:
                    logger.error('Invalid range: {}'.format(item))
                    valid = False
                    break
                m_ports.update(range(s, e + 1))
                continue
            # single number
            elif utils.is_number(item):
                m_ports.add(int(item))
                continue
            else:
                logger.error('Invalid input: {}'.format(item))
                valid = False
                break
        if not valid:
            continue
        out_of_range = []
        for port in m_ports:
            if not port_range_safe(port):
                out_of_range.append(port)
        if out_of_range:
            msg = 'Use port between {} and {}: {}'.format(
                PORT_MININUM,
                PORT_MAXIMUM,
                out_of_range,
            )
            logger.error(msg)
            continue
        if valid and len(m_ports) != m_count:
            q2 = [
                "You type count '{}' at first, ".format(m_count),
                "but now count is '{}'. ".format(len(m_ports)),
                'try again.'
            ]
            logger.error(''.join(q2))
            continue
        if valid:
            break
    m_ports = sorted(list(m_ports))
    logger.info('OK, {}'.format(result))
    return m_ports


def replicas(save, default=None):
    logger.debug('ask replicas')
    deploy_history = config.get_deploy_history()
    if not default:
        default = deploy_history['replicas']
    q = 'How many replicas would you like to create on each master?'
    result = int(askInt(q, default=str(default)))
    if result < 0:
        msg = [
            'The number of master must be greater than or equal to 0.',
            'try again.',
        ]
        logger.error(' '.join(msg))
        return replicas(save, default=default)
    if save:
        deploy_history['replicas'] = result
        config.save_deploy_history(deploy_history)
    logger.info('OK, {}'.format(result))
    return result


def slave_ports(cluster_id, m_count, replicas_count):
    logger.debug('ask slave ports')
    if replicas_count <= 0:
        logger.debug('return empty list')
        return []
    s_count = m_count * replicas_count
    start_s_ports = START_PORT + (cluster_id * MASTER_OFFSET) + SLAVE_OFFSET
    end_s_ports = start_s_ports + s_count - 1
    if start_s_ports == end_s_ports:
        default_s_ports = str(start_s_ports)
    else:
        default_s_ports = '{}-{}'.format(start_s_ports, end_s_ports)
    q = [
        'Please type ports separate with comma(,) ',
        'and use hyphen(-) for range.',
    ]

    while True:
        result = ask(''.join(q), default=default_s_ports)
        result = list(map(lambda x: x.strip(), result.split(',')))
        valid = True
        s_ports = set()
        p = re.compile('[0-9]+-[0-9]+')
        for item in result:
            # range number
            m = p.match(item)
            if m:
                s, e = map(int, item.split('-'))
                if s > e:
                    logger.error('Invalid range: {}'.format(item))
                    valid = False
                    break
                s_ports.update(range(s, e + 1))
                continue
            # single number
            elif utils.is_number(item):
                s_ports.add(int(item))
                continue
            else:
                logger.error('Invalid input: {}'.format(item))
                valid = False
                break
        out_of_range = []
        for port in s_ports:
            if not port_range_safe(port):
                out_of_range.append(port)
        if out_of_range:
            logger.error('Use port between {} and {}: {}'.format(
                PORT_MININUM,
                PORT_MAXIMUM,
                out_of_range
            ))
            continue
        if valid and len(s_ports) != s_count:
            real_replicas_count = len(s_ports) / float(m_count)
            msg = [
                "You type replicas '{}' at first, ".format(replicas),
                "but now count is '{}'. ".format(real_replicas_count),
                'try again.'
            ]
            logger.error(''.join(msg))
            continue
        if valid:
            break
    s_ports = sorted(list(s_ports))
    logger.info('OK, {}'.format(result))
    return s_ports


def ssd_count(save, default=None):
    logger.debug('ask ssd count')
    deploy_history = config.get_deploy_history()
    if not default:
        default = deploy_history['ssd_count']
    q = 'How many ssd would you like to use?'
    result = int(askInt(q, default=str(default)))
    if result <= 0:
        logger.warn("The number of ssd must be greater than 0. try again.")
        return ssd_count(save=save, default=default)
    if save:
        deploy_history['ssd_count'] = result
        config.save_deploy_history(deploy_history)
    logger.info('OK, {}'.format(result))
    return result


def base_directory(default='~/tsr2'):
    logger.debug('ask base directory')
    result = ask('Type base directory for LightningDB', default=default)
    if not result.startswith(('~', '/')):
        logger.error("Invalid path: '{}', try again".format(result))
        return base_directory()
    logger.info('OK, {}'.format(result))
    cli_config = config.get_cli_config()
    cli_config['base_directory'] = result
    config.save_cli_config(cli_config)
    return result


def prefix_of_db_path(save, default=None):
    logger.debug('ask_prefix_of_db_path')
    deploy_history = config.get_deploy_history()
    if not default:
        default = deploy_history['prefix_of_db_path']
    q = 'Type prefix of db path'
    result = ask(q, default=default)
    result = result.strip()
    if save:
        deploy_history['prefix_of_db_path'] = result
        config.save_deploy_history(deploy_history)
    logger.info('OK, {}'.format(result))
    return result


def props(cluster_id, save):
    ret = {}
    ret['hosts'] = hosts(save)
    m_ports = master_ports(save, cluster_id)
    ret['master_ports'] = m_ports
    ret['replicas'] = replicas(save)
    m_count = len(m_ports)
    s_ports = slave_ports(cluster_id, m_count, ret['replicas'])
    ret['slave_ports'] = s_ports
    ret['ssd_count'] = int(ssd_count(save))
    ret['prefix_of_db_path'] = prefix_of_db_path(save)
    return ret


def host_for_monitor(host_list):
    formatted = []
    for i, v in enumerate(host_list):
        formatted.append('    ({}) {}'.format(i + 1, v))
    msg = [
        'Select host',
        '',
        '   [ HOST LIST ]',
        '{}\n'.format('\n'.join(formatted)),
        'Please enter the number you want to see log.',
    ]
    target_num = int(askInt('\n'.join(msg), default='1'))
    while True:
        if target_num > 0 and target_num <= len(host_list):
            break
        msg = [
            'Choose a number ',
            'between 1 and {}'.format(len(host_list)),
            ', please try again'
        ]
        logger.error(''.join(msg))
        target_num = int(askInt(''))
    return host_list[target_num - 1]



from __future__ import print_function

import redis
from .custom_util import PrettySlotGenerator


class CustomClusterNode(object):
    def __init__(self, addr):
        s = addr.split(':')
        if len(s) < 2:
            print('Invalid Addr (given as %s) - use IP:Port format' % addr)
            exit(1)
        ip = s[0]
        port = int(s[1])
        info = {
            'ip': ip,
            'port': port,
            'slots': {},
            'migrating': {},
            'importing': {},
            'replicate': False,
        }
        self.info = info
        self.all_info_list = []
        self.friends = []
        self.dirty = False
        self.r = None
        self.ip = ip
        self.port = port
        self.connect()

    def connect(self):
        self.r = redis.Redis(host=self.ip, port=self.port)
        self.r.ping()

    def to_string(self):
        return '%s:%s' % (self.info['ip'], self.info['port'])

    def info_string(self, info=None):
        if info is None:
            info = self.info
        tag = 'S'
        if self._is_master(info):
            tag = 'M'
        slots = info['slots']
        slot_info = self._get_pretty_slot_info(slots)
        msg = '%s: %s %s %s %s' % (
            tag,
            info['name'],
            info['ip'],
            info['port'],
            slot_info
        )
        return msg

    def load_info(self, o={}):
        conn = self.r
        nodes_res_text = conn.execute_command('cluster nodes')
        info_res_text = conn.execute_command('cluster info')
        self.all_info_list = []
        self.info = {}
        self._load_info(nodes_res_text, info_res_text, o)
        return self

    def get_config_signature(self):
        self.load_info()
        signature = ''
        sorted_info_list = sorted(self.all_info_list, key=lambda x: x['name'])
        for info in sorted_info_list:
            msg = self.info_string(info)
            if msg.startswith('M'):
                signature += (msg + '|')
        return signature

    @staticmethod
    def _get_pretty_slot_info(slots):
        slots = slots
        g = PrettySlotGenerator()
        g.generate(slots)
        return g.to_string()

    def _load_info(self, nodes_res_text, info_res_text, o):
        s = nodes_res_text.split('\n')
        for row in s:
            if len(row) > 0:
                self._load_info_row(row, info_res_text, o)

    def _load_info_row(self, row, info_res_text, o):
        arr = row.split(' ')
        name = arr[0]
        addr = arr[1]
        if '@' in addr:
            addr = addr.split('@')[0]
        s = addr.split(':')
        if len(s) < 2:
            print('Invalid Addr (given as %s) - use IP:Port format' % addr)
            exit(1)
        ip = s[0]
        port = int(s[1])
        flags = arr[2]
        my_master_id = arr[3]
        ping_sent = arr[4]
        ping_recv = arr[5]
        config_epoch = arr[6]
        link_status = arr[7]
        slots = arr[8:]
        info = {
            'name': name,
            'ip': ip,
            'port': port,
            'addr': addr,
            'flags': flags.split(','),
            'replicate': my_master_id == '-',
            'ping_sent': int(ping_sent),
            'ping_recv': int(ping_recv),
            'link_status': link_status,
            'config_epoch': config_epoch,
            'migrating': {},
            'importing': {},
            'my_master_id': my_master_id,
            'slots': {},
        }
        self._load_slots(info, slots)
        self._load_cluster_info_text(info, info_res_text)
        self.all_info_list.append(info)
        if 'myself' in info['flags']:
            self.info = info
        else:
            self.friends.append(info)

    def _is_master(self, info=None):
        if info is None:
            info = self.info
        return 'master' in info['flags']

    @staticmethod
    def _load_cluster_info_text(info, info_res_text):
        items = info_res_text.split('\n')
        for item in items:
            if len(item.strip()) == 0:
                continue
            k, v = item.split(':')
            k = k.strip()
            v = v.strip()
            if k == 'cluster_state':
                info[k] = v
            else:
                info[k] = int(v)

    @staticmethod
    def _load_slots(info, slots):
        for slot in slots:
            if len(slot) == 0:
                continue
            if '->-' in slot:
                slot = slot.replace('[', '').replace(']', '')
                no, dst = slot.split('->-')
                info['migrating'][int(no)] = dst
            elif '-<-' in slot:
                slot = slot.replace('[', '').replace(']', '')
                no, src = slot.split('-<-')
                info['importing'][int(no)] = src
            elif '-' in slot:
                start, stop = slot.split('-')
                CustomClusterNode._add_slots(info, start, stop)
            else:
                CustomClusterNode._add_slots(info, slot, slot)

    @staticmethod
    def _add_slots(info, start, stop):
        for i in range(int(start), int(stop) + 1):
            info['slots'][i] = True



from __future__ import print_function

import errno
import socket
import time
from threading import Thread
import os
import sys
import shutil

import paramiko
import requests

from ltcli import parser
from ltcli.log import logger
from ltcli.exceptions import (
    SSHConnectionError,
    HostConnectionError,
    HostNameError,
    SSHCommandError,
)


def get_ssh(host, port=22):
    """Create SSHClient, connect TCP, and return it

    :param host: host
    :param port: port
    :return: socket (SSHClient)
    """
    try:
        client = paramiko.SSHClient()
        client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        client.load_system_host_keys()
        client.connect(hostname=host, port=port)
        client.hostname = host
        client.port = port
        return client
    except paramiko.ssh_exception.NoValidConnectionsError:
        raise HostConnectionError(host)
    except paramiko.ssh_exception.AuthenticationException:
        raise SSHConnectionError(host)
    except socket.gaierror:
        raise HostNameError(host)


def get_sftp(client):
    """Open sftp

    :param client: SSHClient instance
    :return: opened sftp instance
    """
    try:
        sftp = client.open_sftp()
        return sftp
    except Exception as e:
        logger.debug(e)


def __ssh_execute_async_thread(channel, command):
    channel.exec_command(command)
    while True:
        if channel.exit_status_ready():
            break
        try:
            msg = channel.recv(4096)
            print(msg.decode('utf-8'))
        except socket.error as e:
            err = e.args[0]
            if err == errno.EAGAIN or err == errno.EWOULDBLOCK:
                time.sleep(1)
                continue
            else:
                print(e)
                break


def ssh_execute_async(client, command):
    """Execute ssh asynchronous

    This for for not terminating process (like tail -f)

    :param client: SSHClient
    :param command: command
    """
    channel = client.get_transport().open_session()
    t = Thread(
        target=__ssh_execute_async_thread,
        args=(channel, command,))
    t.daemon = True
    t.start()
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        channel.close()


def ssh_bulk_execute(hosts, command, allow_status=[0]):
    """
    hosts = [
        '192.168.1.1',
        '192.168.1.2',
        '192.168.1.3',
        '192.168.1.4',
    ]
    commmand = 'cp A B'
    """
    def _ssh_exeute_thread(client, command, allow_status, outputs):
        output = ssh_execute(client, command, allow_status)
        outputs.append(output)
    ts = []
    outputs = []
    for host in hosts:
        client = get_ssh(host)
        t = Thread(
            target=_ssh_exeute_thread,
            args=(client, command, allow_status, outputs))
        t.daemon = True
        ts.start()
    for t in ts:
        t.join()
    return outputs


def ssh_execute(client, command, allow_status=[0]):
    """Execute ssh blocking I/O

    We get allow_status as an input.
    We ignore some error status, if it exist in allow_status.

    :param client: SSHClient
    :param command: command
    :param allow_status: list of allow status.
    """
    try:
        logger.debug('[ssh_execute] %s' % command)
        stdin, stdout, stderr = client.exec_command(command)
    except Exception as e:
        print(e)
        raise e
    stdout_msg = ''
    stderr_msg = ''

    stdout_msg = stdout.read()
    stdout_msg = stdout_msg.decode('utf-8')
    if stdout_msg and stdout_msg is not "":
        logger.debug('---------------- command to : %s' % client.hostname)
        logger.debug(command)
        logger.debug('---------------- stdout')
        logger.debug(stdout_msg)

    stderr_msg = stderr.read()
    stderr_msg = stderr_msg.decode('utf-8')
    if stderr_msg and stderr_msg is not "":
        logger.debug('---------------- command to : %s' % client.hostname)
        logger.debug(command)
        logger.debug('---------------- stderr')
        logger.debug(stderr_msg)

    exit_status = stdout.channel.recv_exit_status()
    if exit_status not in allow_status:
        host = client.hostname
        stderr_msg = stderr_msg.encode('utf-8')
        raise SSHCommandError(exit_status, host, stderr_msg)
    return exit_status, stdout_msg, stderr_msg


def is_dir(client, file_path):
    """Determine if directory or not

    :param client: SSHClient
    :param file_path: absolute path of file
    """
    command = "[ -d '{}' ] && echo True || echo False".format(file_path)
    _, stdout, _ = ssh_execute(client, command)
    return stdout.strip() == 'True'


def is_exist(client, file_path):
    """Determine if exist or not

    :param client: SSHClient
    :param file_path: absolute path of file
    """
    command = "[ -e '{}' ] && echo True || echo False".format(file_path)
    _, stdout, _ = ssh_execute(client, command)
    return stdout.strip() == 'True'


def is_exist_files(client, file_path_list):
    """Determine if exist or not

    If there is not any single file, return false.

    :param client: SSHClient
    :param file_path_list: absolute path of file list
    """
    command = []
    for file_path in file_path_list:
        command.append("[ -e '{}' ] &&".format(file_path))
    command.append('echo True || echo False')
    command = ' '.join(command)
    _, stdout, _ = ssh_execute(client, command)
    return stdout.strip() == 'True'


def copy_dir_to_remote(client, local_path, remote_path):
    """copy directory from local to remote
    if already file exist, overwrite file
    copy all files recursively
    directory must exist

    :param client: SSHClient
    :param local_path: absolute path of file
    :param remote_path: absolute path of file
    """
    logger.debug('copy_dir_to_remote')
    logger.debug('copy FROM localhost:{} TO node:{}'.format(
        local_path,
        remote_path
    ))
    sftp = get_sftp(client)
    listdir = os.listdir(local_path)
    for f in listdir:
        r_path = os.path.join(remote_path, f)
        l_path = os.path.join(local_path, f)
        if os.path.isdir(l_path):
            if not is_exist(client, r_path):
                sftp.mkdir(r_path)
            copy_dir_to_remote(client, l_path, r_path)
        else:
            sftp.put(l_path, r_path)


def copy_dir_from_remote(client, remote_path, local_path):
    """copy directory from remote to local

    if already file exist, overwrite file
    copy all files recursively
    directory must exist

    :param client: SSHClient
    :param remote_path: absolute path of file
    :param local_path: absolute path of file
    """
    logger.debug('copy FROM node:{} TO localhost:{}'.format(
        remote_path,
        local_path
    ))
    sftp = get_sftp(client)
    listdir = sftp.listdir(remote_path)
    for f in listdir:
        r_path = os.path.join(remote_path, f)
        l_path = os.path.join(local_path, f)
        if is_dir(client, r_path):
            if not os.path.exists(l_path):
                os.mkdir(l_path)
            copy_dir_from_remote(client, r_path, l_path)
        else:
            sftp.get(r_path, l_path)


def get_home_path(host):
    client = get_ssh(host)
    command = 'echo $HOME'
    _, stdout, _ = ssh_execute(client, command)
    client.close()
    return stdout.strip()


def ping(host, duration=3):
    command = 'ping -c 1 -t {} {} > /dev/null 2>&1'.format(duration, host)
    response = os.system(command)
    print(response)
    logger.debug('ping to {}, respose: {}'.format(host, response))
    if response is not 0:
        raise HostConnectionError(host, status_code=response)


def is_port_empty(host, port):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    result = True
    status = 'OK'
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind((host, port))
    except socket.error as e:
        result = False
        if e.errno == errno.EADDRINUSE:
            status = 'USED'
            logger.debug('{}:{} is already in use'.format(host, port))
        else:
            status = 'FAIL'
            logger.exception(e)
    except Exception as e:
        result = False
        status = 'FAIL'
        logger.exception(e)
    finally:
        return result, status


def get_ip(host):
    try:
        ip = socket.gethostbyname(host)
    except socket.gaierror:
        raise HostNameError(host)
    return ip


def download_file(url, file_path):
    download_path = file_path + '.download'
    file_name = os.path.basename(file_path)
    try:
        with open(download_path, 'wb') as f:
            logger.info('Downloading {}'.format(file_name))
            logger.debug('url: {}'.format(url))
            logger.debug('installer name: {}'.format(file_name))
            response = requests.get(url, stream=True)
            response.raise_for_status()
            total_length = response.headers.get('content-length')
            if total_length is None:
                f.write(response.content)
            else:
                done_length = 0
                total_length = int(total_length)
                for data in response.iter_content(chunk_size=4096):
                    done_length += len(data)
                    f.write(data)
                    done = int(100 * done_length / total_length)
                    comp = '=' * int(done / 2)
                    remain = ' ' * int(50 - int(done / 2))
                    progress = '\r[{}{}] {}%'.format(comp, remain, done)
                    sys.stdout.write(progress)
                    sys.stdout.flush()
            print('')
            shutil.move(download_path, file_path)
            return True
    except requests.exceptions.HTTPError as ex:
        logger.warning(ex)
        return False
    except KeyboardInterrupt as ex:
        print('')
        raise ex
    except BaseException as ex:
        class_name = ex.__class__.__name__
        logger.warning('{}: {}'.format(class_name, url))
        return False
    finally:
        if os.path.isfile(download_path):
            os.remove(download_path)


def get_installers_from_fb_s3(maximum_number=5):
    '''bring up to maximum_number installers in the latest order from s3
    default value of maximum_value is 5
    if there is problem with url or network connection is fail,
    return empty list

    return
    [{
        name:string: file name
        url:string: download url
        type:string: url type
    }]
    '''
    ret = []
    url = 'https://flashbase.s3.ap-northeast-2.amazonaws.com/latest/latest.html'
    warning_msg = "Fail to load installer list from '{}'".format(url)
    try:
        res = requests.get(url)
        status_code = res.status_code
        if status_code >= 400:
            logger.warning('{}:HTTP Status {}'.format(warning_msg, status_code))
        res_text = str(res.text)
        res_text = list(map(lambda x: x.strip(), res_text.split('\n')))
        filtered = list(filter(lambda x: x.startswith('<a href='), res_text))
        for text in filtered:
            if maximum_number <= 0:
                break
            link = parser.get_word_between(text, '<a href="', '">')
            name = parser.get_word_between(text, '<a href=".*">', '/*</a>')
            ret.append({'name': name, 'url': link, 'type': 'download'})
            maximum_number -= 1
        return ret
    except requests.exceptions.ConnectionError:
        logger.warning('{}:Connection failed'.format(warning_msg))
        return []
